# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Entities that are used to describe polls on Discord."""

from __future__ import annotations

__all__: typing.Sequence[str] = ("Poll", "PollAnswer", "PollAnswerCount", "PollLayoutType", "PollMedia", "PollResult")

import typing

import attrs

from hikari.internal import attrs_extensions
from hikari.internal import enums

if typing.TYPE_CHECKING:
    import datetime

    from hikari import emojis


@attrs_extensions.with_copy
@attrs.define(hash=False, kw_only=True, weakref_slot=False)
class PollMedia:
    """Common object backing a poll's questions and answers."""

    text: str | None = attrs.field(default=None, repr=True)
    """The text of the element, or [`None`][] if not present."""

    emoji: emojis.Emoji | None = attrs.field(default=None, repr=True)
    """The emoji of the element, or [`None`][] if not present."""


@attrs_extensions.with_copy
@attrs.define(hash=False, kw_only=True, weakref_slot=False)
class PollAnswer:
    """Represents an answer to a poll."""

    answer_id: int = attrs.field(repr=True)
    """The ID that labels this answer."""

    poll_media: PollMedia = attrs.field(repr=True)
    """The [media][hikari.polls.PollMedia] associated with this answer."""


@attrs_extensions.with_copy
@attrs.define(hash=False, kw_only=True, weakref_slot=False)
class PollResult:
    """Represents a poll result."""

    is_finalized: bool = attrs.field(repr=True)
    """Whether the poll is finalized and the votes are precisely counted."""

    answer_counts: typing.Sequence[PollAnswerCount] = attrs.field(repr=True)
    """The counts for each answer."""


@attrs_extensions.with_copy
@attrs.define(hash=False, kw_only=True, weakref_slot=False)
class PollAnswerCount:
    """Represents the count of a poll answer."""

    id: int = attrs.field(repr=True)
    """The ID of the answer."""

    count: int = attrs.field(repr=True)
    """The number of votes for this answer."""

    me_voted: bool = attrs.field(repr=True)
    """Whether the current user voted for this answer."""


class PollLayoutType(int, enums.Enum):
    """Layout of a poll."""

    DEFAULT = 1
    """The default layout of a poll."""


@attrs_extensions.with_copy
@attrs.define(kw_only=True, repr=True, weakref_slot=False)
class Poll:
    """Represents an existing poll."""

    question: PollMedia = attrs.field(repr=True)
    """The question of the poll."""

    answers: typing.Sequence[PollAnswer] = attrs.field(repr=True)
    """The answers attached to the poll."""

    expiry: datetime.datetime | None = attrs.field(repr=True)
    """The expiry time for the poll."""

    allow_multiselect: bool = attrs.field(repr=True)
    """Whether a user can select multiple answers."""

    layout_type: PollLayoutType = attrs.field(repr=True)
    """The type of layout the poll uses."""

    results: PollResult | None = attrs.field(repr=True)
    """The results of the poll."""



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import typing

import nox

NoxCallbackSigT = typing.Callable[[nox.Session], None]

# Default sessions should be defined here
nox.options.sessions = ["reformat-code", "codespell", "pytest", "ruff", "slotscheck", "mypy", "verify-types"]
nox.options.default_venv_backend = "uv"


def session(**kwargs: typing.Any) -> typing.Callable[[NoxCallbackSigT], NoxCallbackSigT]:  # noqa: ANN401
    """Session wrapper to give default job kwargs."""

    def decorator(func: NoxCallbackSigT) -> NoxCallbackSigT:
        name = func.__name__.replace("_", "-")
        reuse_venv = kwargs.pop("reuse_venv", True)
        return nox.session(name=name, reuse_venv=reuse_venv, **kwargs)(func)

    return decorator


def sync(
    session: nox.Session, /, *, self: bool = False, extras: typing.Sequence[str] = (), groups: typing.Sequence[str] = ()
) -> None:
    """Install session packages using `uv sync`."""
    if extras and not self:
        msg = "When specifying extras, set `self=True`."
        raise RuntimeError(msg)

    args: list[str] = []
    for extra in extras:
        args.extend(("--extra", extra))

    group_flag = "--group" if self else "--only-group"
    for group in groups:
        args.extend((group_flag, group))

    session.run_install(
        "uv", "sync", "--locked", *args, silent=True, env={"UV_PROJECT_ENVIRONMENT": session.virtualenv.location}
    )



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import asyncio
import pathlib
import shutil

import mock
import pytest

from hikari import files
from tests.hikari import hikari_test_helpers


class TestURL:
    def test_default_filename(self):
        url = files.URL("https://i.ytimg.com/vi_webp/dQw4w9WgXcQ/maxresdefault.webp")

        assert url.filename == "maxresdefault.webp"

    def test_set_filename(self):
        url = files.URL("https://i.ytimg.com/vi_webp/dQw4w9WgXcQ/maxresdefault.webp", "yeltsakir.webp")

        assert url.filename == "yeltsakir.webp"


class TestAsyncReaderContextManager:
    @pytest.fixture
    def reader(self):
        return hikari_test_helpers.mock_class_namespace(files.AsyncReaderContextManager)

    def test___enter__(self, reader):
        with pytest.raises(TypeError, match=" is async-only, did you mean 'async with'?"):
            reader().__enter__()

    def test___exit__(self, reader):
        try:
            reader().__exit__(None, None, None)
        except AttributeError as exc:
            pytest.fail(exc)


def test__open_read_path():
    expanded_path = mock.Mock()
    path = mock.Mock(expanduser=mock.Mock(return_value=expanded_path))

    assert files._open_read_path(path) is expanded_path.open.return_value

    expanded_path.open.assert_called_once_with("rb")


class TestThreadedFileReaderContextManagerImpl:
    @pytest.mark.asyncio
    async def test_enter_dunder_method_when_already_open(self):
        manager = files._ThreadedFileReaderContextManagerImpl(mock.Mock(), "ea", pathlib.Path("ea"))
        manager.file = mock.Mock()
        with pytest.raises(RuntimeError, match="File is already open"):
            await manager.__aenter__()

    @pytest.mark.asyncio
    async def test_exit_dunder_method_when_not_open(self):
        manager = files._ThreadedFileReaderContextManagerImpl(mock.Mock(), "ea", pathlib.Path("ea"))

        with pytest.raises(RuntimeError, match="File isn't open"):
            await manager.__aexit__(None, None, None)

    @pytest.mark.asyncio
    async def test_context_manager(self):
        mock_file = mock.Mock()
        executor = object()
        path = pathlib.Path("test/path/")

        loop = mock.Mock(run_in_executor=mock.AsyncMock(side_effect=[mock_file, None]))

        context_manager = files._ThreadedFileReaderContextManagerImpl(executor, "meow.txt", path)

        with mock.patch.object(asyncio, "get_running_loop", return_value=loop):
            async with context_manager as reader:
                assert context_manager.file is mock_file

                assert reader.filename == "meow.txt"
                assert reader._executor is executor

                loop.run_in_executor.assert_called_once_with(executor, files._open_read_path, path)
                loop.run_in_executor.reset_mock()

        loop.run_in_executor.assert_called_once_with(executor, mock_file.close)
        assert context_manager.file is None


class TestToWritePath:
    def test_when_dir(self):
        mock_path = mock.Mock(is_dir=mock.Mock(return_value=True))
        joinpath = mock_path.joinpath.return_value = mock.Mock(exists=mock.Mock(return_value=False))

        with mock.patch.object(files, "ensure_path", return_value=mock_path) as ensure_path:
            assert (
                files._to_write_path("test_path", "some_filename.png", force=False) is joinpath.expanduser.return_value
            )

        mock_path.joinpath.assert_called_once_with("some_filename.png")
        ensure_path.assert_called_once_with("test_path")

    def test_when_exists_but_not_force(self):
        mock_path = mock.Mock(is_dir=mock.Mock(return_value=False), exists=mock.Mock(return_value=True))

        with mock.patch.object(files, "ensure_path", return_value=mock_path) as ensure_path:
            with pytest.raises(FileExistsError):
                files._to_write_path("test_path", "some_filename.png", force=False)

        ensure_path.assert_called_once_with("test_path")

    def test_when_exists_but_force(self):
        mock_path = mock.Mock(is_dir=mock.Mock(return_value=False), exists=mock.Mock(return_value=True))

        with mock.patch.object(files, "ensure_path", return_value=mock_path) as ensure_path:
            assert (
                files._to_write_path("test_path", "some_filename.png", force=True) is mock_path.expanduser.return_value
            )

        ensure_path.assert_called_once_with("test_path")


def test_open_write_path():
    with mock.patch.object(files, "_to_write_path") as to_write_path:
        assert (
            files._open_write_path("path", "some_filename.png", False) is to_write_path.return_value.open.return_value
        )

    to_write_path.assert_called_once_with("path", "some_filename.png", force=False)
    to_write_path.return_value.open.assert_called_once_with("wb")


class TestResource:
    @pytest.fixture
    def resource(self):
        class MockReader:
            data = iter(("never", "gonna", "give", "you", "up"))

            async def __aenter__(self):
                return self

            async def __aexit__(self, *args, **kwargs):
                return

            def __aiter__(self):
                return self

            async def __anext__(self):
                try:
                    return next(self.data)
                except StopIteration:
                    raise StopAsyncIteration from None

        class ResourceImpl(files.Resource):
            stream = mock.Mock(return_value=MockReader())
            url = "https://myspace.com/rickastley/lyrics.txt"
            filename = "lyrics.txt"

        return ResourceImpl()

    @pytest.mark.asyncio
    async def test_save(self, resource):
        executor = object()
        file_open = mock.Mock()
        file_open.write = mock.Mock()
        loop = mock.Mock(run_in_executor=mock.AsyncMock(side_effect=[file_open, None, None, None, None, None, None]))

        with mock.patch.object(asyncio, "get_running_loop", return_value=loop):
            await resource.save("rickroll/lyrics.txt", executor=executor, force=True)

        assert loop.run_in_executor.call_count == 7
        loop.run_in_executor.assert_has_calls(
            [
                mock.call(executor, files._open_write_path, "rickroll/lyrics.txt", "lyrics.txt", True),
                mock.call(executor, file_open.write, "never"),
                mock.call(executor, file_open.write, "gonna"),
                mock.call(executor, file_open.write, "give"),
                mock.call(executor, file_open.write, "you"),
                mock.call(executor, file_open.write, "up"),
                mock.call(executor, file_open.close),
            ]
        )


def test_copy_to_path():
    with mock.patch.object(files, "_to_write_path") as to_write_path:
        with mock.patch.object(shutil, "copy2") as copy2:
            files._copy_to_path("original_path", "path", "some_filename.png", False)

    to_write_path.assert_called_once_with("path", "some_filename.png", force=False)
    copy2.assert_called_once_with("original_path", to_write_path.return_value)


class TestFile:
    @pytest.fixture
    def file_obj(self):
        return files.File("one/path/something.txt")

    @pytest.mark.asyncio
    async def test_save(self, file_obj):
        mock_executor = object()
        loop = mock.Mock(run_in_executor=mock.AsyncMock())

        with mock.patch.object(asyncio, "get_running_loop", return_value=loop):
            with mock.patch.object(files.Resource, "save") as super_save:
                await file_obj.save("some_path/", executor=mock_executor, force=True)

        super_save.assert_not_called()
        loop.run_in_executor.assert_awaited_once_with(
            mock_executor,
            files._copy_to_path,
            pathlib.Path("one/path/something.txt"),
            "some_path/",
            "something.txt",
            True,
        )


def test_write_bytes():
    with mock.patch.object(files, "_to_write_path") as to_write_path:
        files._write_bytes("path", "some_filename.png", b"some bytes", force=False)

    to_write_path.assert_called_once_with("path", "some_filename.png", force=False)
    to_write_path.return_value.write_bytes.assert_called_once_with(b"some bytes")


class TestBytes:
    @pytest.fixture
    def bytes_obj(self):
        return files.Bytes(b"some data", "something.txt")

    @pytest.mark.parametrize("data_type", [bytes, bytearray, memoryview])
    @pytest.mark.asyncio
    async def test_save(self, bytes_obj, data_type):
        bytes_obj.data = mock.Mock(data_type)
        mock_executor = object()
        loop = mock.Mock(run_in_executor=mock.AsyncMock())

        with mock.patch.object(asyncio, "get_running_loop", return_value=loop):
            with mock.patch.object(files.Resource, "save") as super_save:
                await bytes_obj.save("some_path/", executor=mock_executor, force=True)

        super_save.assert_not_called()
        loop.run_in_executor.assert_awaited_once_with(
            mock_executor, files._write_bytes, "some_path/", "something.txt", bytes_obj.data, True
        )

    @pytest.mark.asyncio
    async def test_save_when_data_is_not_bytes(self, bytes_obj):
        bytes_obj.data = object()
        mock_executor = object()

        with mock.patch.object(asyncio, "get_running_loop") as get_running_loop:
            with mock.patch.object(files.Resource, "save") as super_save:
                await bytes_obj.save("some_path/", executor=mock_executor, force=True)

        super_save.assert_called_once_with("some_path/", executor=mock_executor, force=True)
        get_running_loop.return_value.run_in_executor.assert_not_called()



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import datetime
import typing

import mock
import pytest

from hikari import auto_mod
from hikari import channels as channel_models
from hikari import emojis as emoji_models
from hikari import traits
from hikari import undefined
from hikari import users as user_models
from hikari.api import shard
from hikari.events import application_events
from hikari.events import auto_mod_events
from hikari.events import channel_events
from hikari.events import guild_events
from hikari.events import interaction_events
from hikari.events import lifetime_events
from hikari.events import member_events
from hikari.events import message_events
from hikari.events import monetization_events
from hikari.events import poll_events
from hikari.events import reaction_events
from hikari.events import role_events
from hikari.events import scheduled_events
from hikari.events import shard_events
from hikari.events import stage_events
from hikari.events import typing_events
from hikari.events import user_events
from hikari.events import voice_events
from hikari.impl import event_factory as event_factory_
from hikari.interactions import base_interactions


class TestEventFactoryImpl:
    @pytest.fixture
    def mock_app(self):
        return mock.Mock(traits.RESTAware)

    @pytest.fixture
    def mock_shard(self):
        return mock.Mock(shard.GatewayShard)

    @pytest.fixture
    def event_factory(self, mock_app):
        return event_factory_.EventFactoryImpl(mock_app)

    ######################
    # APPLICATION EVENTS #
    ######################

    def test_deserialize_application_command_permission_update_event(self, event_factory, mock_app, mock_shard):
        mock_payload = object()

        event = event_factory.deserialize_application_command_permission_update_event(mock_shard, mock_payload)

        mock_app.entity_factory.deserialize_guild_command_permissions.assert_called_once_with(mock_payload)
        assert isinstance(event, application_events.ApplicationCommandPermissionsUpdateEvent)
        assert event.app is mock_app
        assert event.shard is mock_shard
        assert event.permissions is mock_app.entity_factory.deserialize_guild_command_permissions.return_value

    ##################
    # CHANNEL EVENTS #
    ##################

    def test_deserialize_guild_channel_create_event(self, event_factory, mock_app, mock_shard):
        mock_app.entity_factory.deserialize_channel.return_value = mock.Mock(
            spec=channel_models.PermissibleGuildChannel
        )
        mock_payload = mock.Mock(app=mock_app)

        event = event_factory.deserialize_guild_channel_create_event(mock_shard, mock_payload)

        mock_app.entity_factory.deserialize_channel.assert_called_once_with(mock_payload)
        assert isinstance(event, channel_events.GuildChannelCreateEvent)
        assert event.shard is mock_shard
        assert event.channel is mock_app.entity_factory.deserialize_channel.return_value

    def test_deserialize_guild_channel_update_event(self, event_factory, mock_app, mock_shard):
        mock_app.entity_factory.deserialize_channel.return_value = mock.Mock(
            spec=channel_models.PermissibleGuildChannel
        )
        mock_old_channel = object()
        mock_payload = object()

        event = event_factory.deserialize_guild_channel_update_event(
            mock_shard, mock_payload, old_channel=mock_old_channel
        )

        mock_app.entity_factory.deserialize_channel.assert_called_once_with(mock_payload)
        assert isinstance(event, channel_events.GuildChannelUpdateEvent)
        assert event.shard is mock_shard
        assert event.channel is mock_app.entity_factory.deserialize_channel.return_value
        assert event.old_channel is mock_old_channel

    def test_deserialize_guild_channel_delete_event(self, event_factory, mock_app, mock_shard):
        mock_app.entity_factory.deserialize_channel.return_value = mock.Mock(
            spec=channel_models.PermissibleGuildChannel
        )
        mock_payload = mock.Mock(app=mock_app)

        event = event_factory.deserialize_guild_channel_delete_event(mock_shard, mock_payload)

        mock_app.entity_factory.deserialize_channel.assert_called_once_with(mock_payload)
        assert isinstance(event, channel_events.GuildChannelDeleteEvent)
        assert event.shard is mock_shard
        assert event.channel is mock_app.entity_factory.deserialize_channel.return_value

    def test_deserialize_channel_pins_update_event_for_guild(self, event_factory, mock_app, mock_shard):
        mock_payload = {"channel_id": "123435", "last_pin_timestamp": None, "guild_id": "43123123"}

        event = event_factory.deserialize_channel_pins_update_event(mock_shard, mock_payload)

        assert isinstance(event, channel_events.GuildPinsUpdateEvent)
        assert event.app is mock_app
        assert event.shard is mock_shard
        assert event.guild_id == 43123123
        assert event.last_pin_timestamp is None

    def test_deserialize_channel_pins_update_event_for_dm(self, event_factory, mock_app, mock_shard):
        mock_payload = {"channel_id": "123435", "last_pin_timestamp": "2020-03-15T15:23:32.686000+00:00"}

        event = event_factory.deserialize_channel_pins_update_event(mock_shard, mock_payload)

        assert isinstance(event, channel_events.DMPinsUpdateEvent)
        assert event.app is mock_app
        assert event.shard is mock_shard
        assert event.last_pin_timestamp == datetime.datetime(
            2020, 3, 15, 15, 23, 32, 686000, tzinfo=datetime.timezone.utc
        )

    def test_deserialize_channel_pins_update_event_without_last_pin_timestamp(
        self, event_factory, mock_app, mock_shard
    ):
        mock_payload = {"channel_id": "123435", "guild_id": "43123123"}

        event = event_factory.deserialize_channel_pins_update_event(mock_shard, mock_payload)

        assert event.last_pin_timestamp is None

    def test_deserialize_guild_thread_create_event(
        self, event_factory: event_factory_.EventFactoryImpl, mock_app: mock.Mock, mock_shard: shard.GatewayShard
    ):
        mock_payload = mock.Mock()

        event = event_factory.deserialize_guild_thread_create_event(mock_shard, mock_payload)

        assert event.shard is mock_shard
        assert event.thread is mock_app.entity_factory.deserialize_guild_thread.return_value
        mock_app.entity_factory.deserialize_guild_thread.assert_called_once_with(mock_payload)
        assert isinstance(event, channel_events.GuildThreadCreateEvent)

    def test_deserialize_guild_thread_access_event(
        self, event_factory: event_factory_.EventFactoryImpl, mock_app: mock.Mock, mock_shard: shard.GatewayShard
    ):
        mock_payload = mock.Mock()

        event = event_factory.deserialize_guild_thread_access_event(mock_shard, mock_payload)

        assert event.shard is mock_shard
        assert event.thread is mock_app.entity_factory.deserialize_guild_thread.return_value
        mock_app.entity_factory.deserialize_guild_thread.assert_called_once_with(mock_payload)
        assert isinstance(event, channel_events.GuildThreadAccessEvent)

    def test_deserialize_guild_thread_update_event(
        self, event_factory: event_factory_.EventFactoryImpl, mock_app: mock.Mock, mock_shard: shard.GatewayShard
    ):
        mock_old_thread = object()
        mock_payload = mock.Mock()

        event = event_factory.deserialize_guild_thread_update_event(
            mock_shard, mock_payload, old_thread=mock_old_thread
        )

        mock_app.entity_factory.deserialize_guild_thread.assert_called_once_with(mock_payload)
        assert isinstance(event, channel_events.GuildThreadUpdateEvent)
        assert event.shard is mock_shard
        assert event.thread is mock_app.entity_factory.deserialize_guild_thread.return_value
        assert event.old_thread is mock_old_thread

    def test_deserialize_guild_thread_delete_event(
        self, event_factory: event_factory_.EventFactoryImpl, mock_app: mock.Mock, mock_shard: shard.GatewayShard
    ):
        mock_payload = {"id": "12332123321", "guild_id": "54544234342", "parent_id": "9494949", "type": 11}

        event = event_factory.deserialize_guild_thread_delete_event(mock_shard, mock_payload)

        assert event.app is mock_app
        assert event.shard is mock_shard
        assert event.thread_id == 12332123321
        assert event.guild_id == 54544234342
        assert event.parent_id == 9494949
        assert event.type is channel_models.ChannelType.GUILD_PUBLIC_THREAD
        assert isinstance(event, channel_events.GuildThreadDeleteEvent)

    def test_deserialize_thread_members_update_event(
        self, event_factory: event_factory_.EventFactoryImpl, mock_app: mock.Mock, mock_shard: shard.GatewayShard
    ):
        mock_thread_member_payload = {"id": "393939393", "user_id": "3933993"}
        mock_other_thread_member_payload = {"id": "393994954", "user_id": "123321123"}
        mock_thread_member = mock.Mock(user_id=123321123)
        mock_other_thread_member = mock.Mock(user_id=5454234)
        mock_app.entity_factory.deserialize_thread_member.side_effect = [mock_thread_member, mock_other_thread_member]
        payload = {
            "id": "92929929",
            "guild_id": "92929292",
            "member_count": "32",
            "added_members": [mock_thread_member_payload, mock_other_thread_member_payload],
            "removed_member_ids": ["4949534", "123321", "54234"],
        }

        event = event_factory.deserialize_thread_members_update_event(mock_shard, payload)

        assert event.app is mock_app
        assert event.shard is mock_shard
        assert event.thread_id == 92929929
        assert event.guild_id == 92929292
        assert event.added_members == {123321123: mock_thread_member, 5454234: mock_other_thread_member}
        assert event.removed_member_ids == [4949534, 123321, 54234]
        assert event.guild_members == {}
        assert event.guild_presences == {}
        mock_app.entity_factory.deserialize_thread_member.assert_has_calls(
            [mock.call(mock_thread_member_payload), mock.call(mock_other_thread_member_payload)]
        )

    def test_deserialize_thread_members_update_event_when_presences_and_real_members(
        self, event_factory: event_factory_.EventFactoryImpl, mock_app: mock.Mock, mock_shard: shard.GatewayShard
    ):
        mock_presence_payload = mock.Mock()
        mock_other_presence_payload = mock.Mock()
        mock_guild_member_payload = mock.Mock()
        mock_other_guild_member_payload = mock.Mock()
        mock_thread_member_payload = {
            "id": "393939393",
            "user_id": "3933993",
            "member": mock_guild_member_payload,
            "presence": mock_presence_payload,
        }
        mock_other_thread_member_payload = {
            "id": "393994954",
            "user_id": "123321123",
            "member": mock_other_guild_member_payload,
            "presence": mock_other_presence_payload,
        }
        mock_thread_member = mock.Mock(user_id=3933993)
        mock_other_thread_member = mock.Mock(user_id=123321123)
        mock_presence = mock.Mock()
        mock_other_presence = mock.Mock()
        mock_guild_member = mock.Mock()
        mock_other_guild_member = mock.Mock()
        mock_app.entity_factory.deserialize_thread_member.side_effect = [mock_thread_member, mock_other_thread_member]
        mock_app.entity_factory.deserialize_member.side_effect = [mock_guild_member, mock_other_guild_member]
        mock_app.entity_factory.deserialize_member_presence.side_effect = [mock_presence, mock_other_presence]
        payload = {
            "id": "92929929",
            "guild_id": "123321123123",
            "member_count": "32",
            "added_members": [mock_thread_member_payload, mock_other_thread_member_payload],
            "removed_member_ids": ["4949534", "123321", "54234"],
        }

        event = event_factory.deserialize_thread_members_update_event(mock_shard, payload)

        assert event.app is mock_app
        assert event.shard is mock_shard
        assert event.thread_id == 92929929
        assert event.guild_id == 123321123123
        assert event.added_members == {3933993: mock_thread_member, 123321123: mock_other_thread_member}
        assert event.removed_member_ids == [4949534, 123321, 54234]
        assert event.guild_members == {3933993: mock_guild_member, 123321123: mock_other_guild_member}
        assert event.guild_presences == {3933993: mock_presence, 123321123: mock_other_presence}
        mock_app.entity_factory.deserialize_thread_member.assert_has_calls(
            [mock.call(mock_thread_member_payload), mock.call(mock_other_thread_member_payload)]
        )
        mock_app.entity_factory.deserialize_member.assert_has_calls(
            [
                mock.call(mock_guild_member_payload, guild_id=123321123123),
                mock.call(mock_other_guild_member_payload, guild_id=123321123123),
            ]
        )
        mock_app.entity_factory.deserialize_member_presence.assert_has_calls(
            [
                mock.call(mock_presence_payload, guild_id=123321123123),
                mock.call(mock_other_presence_payload, guild_id=123321123123),
            ]
        )

    def test_deserialize_thread_members_update_event_partial(
        self, event_factory: event_factory_.EventFactoryImpl, mock_shard: shard.GatewayShard
    ):
        payload = {"id": "92929929", "guild_id": "92929292", "member_count": "32"}

        event = event_factory.deserialize_thread_members_update_event(mock_shard, payload)

        assert event.added_members == {}
        assert event.removed_member_ids == []
        assert event.guild_members == {}
        assert event.guild_presences == {}

    def test_deserialize_thread_list_sync_event(
        self, event_factory: event_factory_.EventFactoryImpl, mock_app: mock.Mock, mock_shard: shard.GatewayShard
    ):
        mock_thread_payload = {"id": "342123123", "name": "nyaa"}
        mock_other_thread_payload = {"id": "5454123123", "name": "meow"}
        mock_not_in_thread_payload = {"id": "94949494", "name": "123123123"}
        mokc_member_payload = {"id": "342123123", "user_id": "9349393939"}
        mock_other_member_payload = {"id": "5454123123", "user_id": "34949499494"}
        mock_thread = mock.Mock(id=342123123)
        mock_other_thread = mock.Mock(id=5454123123)
        mock_not_in_thread = mock.Mock(id=94949494)
        mock_member = mock.Mock(thread_id=342123123)
        mock_other_member = mock.Mock(thread_id=5454123123)
        mock_app.entity_factory.deserialize_guild_thread.side_effect = [
            mock_thread,
            mock_not_in_thread,
            mock_other_thread,
        ]
        mock_app.entity_factory.deserialize_thread_member.side_effect = [mock_member, mock_other_member]
        mock_payload = {
            "guild_id": "43123123",
            "channel_ids": ["54123", "123431", "43939", "12343123"],
            "threads": [mock_thread_payload, mock_not_in_thread_payload, mock_other_thread_payload],
            "members": [mock_other_member_payload, mokc_member_payload],
        }

        event = event_factory.deserialize_thread_list_sync_event(mock_shard, mock_payload)

        assert event.app is mock_app
        assert event.shard is mock_shard
        assert event.guild_id == 43123123
        assert event.channel_ids == [54123, 123431, 43939, 12343123]
        assert event.threads == {342123123: mock_thread, 5454123123: mock_other_thread, 94949494: mock_not_in_thread}
        mock_app.entity_factory.deserialize_thread_member.assert_has_calls(
            [mock.call(mock_other_member_payload), mock.call(mokc_member_payload)]
        )
        mock_app.entity_factory.deserialize_guild_thread.assert_has_calls(
            [
                mock.call(mock_thread_payload, guild_id=43123123, member=mock_member),
                mock.call(mock_not_in_thread_payload, guild_id=43123123, member=None),
                mock.call(mock_other_thread_payload, guild_id=43123123, member=mock_other_member),
            ]
        )

    def test_deserialize_thread_list_sync_event_when_not_channel_ids(
        self, event_factory: event_factory_.EventFactoryImpl, mock_app: mock.Mock, mock_shard: shard.GatewayShard
    ):
        mock_payload = {"guild_id": "123321", "threads": [], "members": []}

        event = event_factory.deserialize_thread_list_sync_event(mock_shard, mock_payload)

        assert event.channel_ids is None

    def test_deserialize_webhook_update_event(self, event_factory, mock_app, mock_shard):
        mock_payload = {"guild_id": "123123", "channel_id": "4393939"}

        event = event_factory.deserialize_webhook_update_event(mock_shard, mock_payload)

        assert isinstance(event, channel_events.WebhookUpdateEvent)
        assert event.app is mock_app
        assert event.shard is mock_shard
        assert event.channel_id == 4393939
        assert event.guild_id == 123123

    def test_deserialize_invite_create_event(self, event_factory, mock_app, mock_shard):
        mock_payload = mock.Mock(app=mock_app)

        event = event_factory.deserialize_invite_create_event(mock_shard, mock_payload)

        mock_app.entity_factory.deserialize_invite_with_metadata.assert_called_once_with(mock_payload)
        assert isinstance(event, channel_events.InviteCreateEvent)
        assert event.shard is mock_shard
        assert event.invite is mock_app.entity_factory.deserialize_invite_with_metadata.return_value

    def test_deserialize_invite_delete_event(self, event_factory, mock_app, mock_shard):
        mock_payload = {"guild_id": "1231234", "channel_id": "123123", "code": "no u"}
        mock_old_invite = object()

        event = event_factory.deserialize_invite_delete_event(mock_shard, mock_payload, old_invite=mock_old_invite)

        assert isinstance(event, channel_events.InviteDeleteEvent)
        assert event.app is mock_app
        assert event.shard is mock_shard
        assert event.channel_id == 123123
        assert event.guild_id == 1231234
        assert event.code == "no u"
        assert event.old_invite is mock_old_invite

    #################
    # TYPING EVENTS #
    ##################

    def test_deserialize_typing_start_event_for_guild(self, event_factory, mock_app, mock_shard):
        mock_member_payload = object()
        mock_payload = {
            "guild_id": "123321",
            "channel_id": "48585858",
            "timestamp": 7634521233,
            "member": mock_member_payload,
        }
        mock_app.entity_factory.deserialize_member.return_value = mock.Mock(app=mock_app)

        event = event_factory.deserialize_typing_start_event(mock_shard, mock_payload)

        mock_app.entity_factory.deserialize_member.assert_called_once_with(mock_member_payload, guild_id=123321)
        assert isinstance(event, typing_events.GuildTypingEvent)
        assert event.shard is mock_shard
        assert event.channel_id == 48585858
        assert event.guild_id == 123321
        assert event.timestamp == datetime.datetime(2211, 12, 6, 12, 20, 33, tzinfo=datetime.timezone.utc)
        assert event.member == mock_app.entity_factory.deserialize_member.return_value

    def test_deserialize_typing_start_event_for_dm(self, event_factory, mock_app, mock_shard):
        mock_payload = {"channel_id": "534123", "timestamp": 7634521212, "user_id": "9494994"}

        event = event_factory.deserialize_typing_start_event(mock_shard, mock_payload)

        assert isinstance(event, typing_events.DMTypingEvent)
        assert event.app is mock_app
        assert event.shard is mock_shard
        assert event.channel_id == 534123
        assert event.timestamp == datetime.datetime(2211, 12, 6, 12, 20, 12, tzinfo=datetime.timezone.utc)
        assert event.user_id == 9494994

    ################
    # GUILD EVENTS #
    ################

    def test_deserialize_guild_available_event(self, event_factory, mock_app, mock_shard):
        mock_payload = mock.Mock(app=mock_app)

        event = event_factory.deserialize_guild_available_event(mock_shard, mock_payload)

        mock_app.entity_factory.deserialize_gateway_guild.assert_called_once_with(
            mock_payload, user_id=mock_shard.get_user_id.return_value
        )
        assert isinstance(event, guild_events.GuildAvailableEvent)
        assert event.shard is mock_shard
        guild_definition = mock_app.entity_factory.deserialize_gateway_guild.return_value
        assert event.guild is guild_definition.guild.return_value
        assert event.emojis is guild_definition.emojis.return_value
        assert event.stickers is guild_definition.stickers.return_value
        assert event.roles is guild_definition.roles.return_value
        assert event.channels is guild_definition.channels.return_value
        assert event.members is guild_definition.members.return_value
        assert event.presences is guild_definition.presences.return_value
        assert event.voice_states is guild_definition.voice_states.return_value
        guild_definition.guild.assert_called_once_with()
        guild_definition.emojis.assert_called_once_with()
        guild_definition.stickers.assert_called_once_with()
        guild_definition.roles.assert_called_once_with()
        guild_definition.channels.assert_called_once_with()
        guild_definition.members.assert_called_once_with()
        guild_definition.presences.assert_called_once_with()
        guild_definition.voice_states.assert_called_once_with()
        mock_shard.get_user_id.assert_called_once_with()

    def test_deserialize_guild_join_event(self, event_factory, mock_app, mock_shard):
        mock_payload = mock.Mock(app=mock_app)

        event = event_factory.deserialize_guild_join_event(mock_shard, mock_payload)

        mock_app.entity_factory.deserialize_gateway_guild.assert_called_once_with(
            mock_payload, user_id=mock_shard.get_user_id.return_value
        )
        assert isinstance(event, guild_events.GuildJoinEvent)
        assert event.shard is mock_shard
        guild_definition = mock_app.entity_factory.deserialize_gateway_guild.return_value
        assert event.guild is guild_definition.guild.return_value
        assert event.emojis is guild_definition.emojis.return_value
        assert event.roles is guild_definition.roles.return_value
        assert event.channels is guild_definition.channels.return_value
        assert event.members is guild_definition.members.return_value
        assert event.presences is guild_definition.presences.return_value
        assert event.voice_states is guild_definition.voice_states.return_value
        mock_shard.get_user_id.assert_called_once_with()

    def test_deserialize_guild_update_event(self, event_factory, mock_app, mock_shard):
        mock_payload = mock.Mock(app=mock_app)
        mock_old_guild = object()

        event = event_factory.deserialize_guild_update_event(mock_shard, mock_payload, old_guild=mock_old_guild)

        mock_app.entity_factory.deserialize_gateway_guild.assert_called_once_with(
            mock_payload, user_id=mock_shard.get_user_id.return_value
        )
        assert isinstance(event, guild_events.GuildUpdateEvent)
        assert event.shard is mock_shard
        guild_definition = mock_app.entity_factory.deserialize_gateway_guild.return_value
        assert event.guild is guild_definition.guild.return_value
        assert event.emojis is guild_definition.emojis.return_value
        assert event.roles is guild_definition.roles.return_value
        assert event.old_guild is mock_old_guild
        guild_definition.guild.assert_called_once_with()
        guild_definition.emojis.assert_called_once_with()
        guild_definition.roles.assert_called_once_with()
        mock_shard.get_user_id.assert_called_once_with()

    def test_deserialize_guild_leave_event(self, event_factory, mock_app, mock_shard):
        mock_payload = {"id": "43123123"}
        mock_old_guild = object()

        event = event_factory.deserialize_guild_leave_event(mock_shard, mock_payload, old_guild=mock_old_guild)

        assert isinstance(event, guild_events.GuildLeaveEvent)
        assert event.app is mock_app
        assert event.shard is mock_shard
        assert event.guild_id == 43123123
        assert event.old_guild is mock_old_guild

    def test_deserialize_guild_unavailable_event(self, event_factory, mock_app, mock_shard):
        mock_payload = {"id": "6541233"}

        event = event_factory.deserialize_guild_unavailable_event(mock_shard, mock_payload)

        assert isinstance(event, guild_events.GuildUnavailableEvent)
        assert event.app is mock_app
        assert event.shard is mock_shard
        assert event.guild_id == 6541233

    def test_deserialize_guild_ban_add_event(self, event_factory, mock_app, mock_shard):
        mock_user_payload = mock.Mock(app=mock_app)
        mock_payload = {"guild_id": "4212312", "user": mock_user_payload}

        event = event_factory.deserialize_guild_ban_add_event(mock_shard, mock_payload)

        mock_app.entity_factory.deserialize_user.assert_called_once_with(mock_user_payload)
        assert isinstance(event, guild_events.BanCreateEvent)
        assert event.shard is mock_shard
        assert event.guild_id == 4212312
        assert event.user is mock_app.entity_factory.deserialize_user.return_value

    def test_deserialize_guild_ban_remove_event(self, event_factory, mock_app, mock_shard):
        mock_user_payload = mock.Mock(app=mock_app)
        mock_payload = {"guild_id": "9292929", "user": mock_user_payload}

        event = event_factory.deserialize_guild_ban_remove_event(mock_shard, mock_payload)

        mock_app.entity_factory.deserialize_user.assert_called_once_with(mock_user_payload)
        assert isinstance(event, guild_events.BanDeleteEvent)
        assert event.shard is mock_shard
        assert event.guild_id == 9292929
        assert event.user is mock_app.entity_factory.deserialize_user.return_value

    def test_deserialize_guild_emojis_update_event(self, event_factory, mock_app, mock_shard):
        mock_emoji_payload = object()
        mock_old_emojis = object()
        mock_payload = {"guild_id": "123431", "emojis": [mock_emoji_payload]}

        event = event_factory.deserialize_guild_emojis_update_event(
            mock_shard, mock_payload, old_emojis=mock_old_emojis
        )

        mock_app.entity_factory.deserialize_known_custom_emoji.assert_called_once_with(
            mock_emoji_payload, guild_id=123431
        )
        assert isinstance(event, guild_events.EmojisUpdateEvent)
        assert event.app is mock_app
        assert event.shard is mock_shard
        assert event.emojis == [mock_app.entity_factory.deserialize_known_custom_emoji.return_value]
        assert event.guild_id == 123431
        assert event.old_emojis is mock_old_emojis

    def test_deserialize_guild_stickers_update_event(self, event_factory, mock_app, mock_shard):
        mock_sticker_payload = object()
        mock_old_stickers = object()
        mock_payload = {"guild_id": "472", "stickers": [mock_sticker_payload]}

        event = event_factory.deserialize_guild_stickers_update_event(
            mock_shard, mock_payload, old_stickers=mock_old_stickers
        )

        mock_app.entity_factory.deserialize_guild_sticker.assert_called_once_with(mock_sticker_payload)
        assert isinstance(event, guild_events.StickersUpdateEvent)
        assert event.app is mock_app
        assert event.shard is mock_shard
        assert event.stickers == [mock_app.entity_factory.deserialize_guild_sticker.return_value]
        assert event.guild_id == 472
        assert event.old_stickers is mock_old_stickers

    def test_deserialize_integration_create_event(self, event_factory, mock_app, mock_shard):
        mock_payload = object()

        event = event_factory.deserialize_integration_create_event(mock_shard, mock_payload)

        mock_app.entity_factory.deserialize_integration.assert_called_once_with(mock_payload)
        assert isinstance(event, guild_events.IntegrationCreateEvent)
        assert event.app is mock_app
        assert event.shard is mock_shard
        assert event.integration is mock_app.entity_factory.deserialize_integration.return_value

    def test_deserialize_integration_delete_event_with_application_id(self, event_factory, mock_app, mock_shard):
        mock_payload = {"id": "123321", "guild_id": "59595959", "application_id": "934949494"}

        event = event_factory.deserialize_integration_delete_event(mock_shard, mock_payload)

        assert isinstance(event, guild_events.IntegrationDeleteEvent)
        assert event.app is mock_app
        assert event.shard is mock_shard
        assert event.id == 123321
        assert event.guild_id == 59595959
        assert event.application_id == 934949494

    def test_deserialize_integration_delete_event_without_application_id(self, event_factory, mock_app, mock_shard):
        mock_payload = {"id": "123321", "guild_id": "59595959"}

        event = event_factory.deserialize_integration_delete_event(mock_shard, mock_payload)

        assert event.application_id is None

    def test_deserialize_integration_update_event(self, event_factory, mock_app, mock_shard):
        mock_payload = object()

        event = event_factory.deserialize_integration_update_event(mock_shard, mock_payload)

        mock_app.entity_factory.deserialize_integration.assert_called_once_with(mock_payload)
        assert isinstance(event, guild_events.IntegrationUpdateEvent)
        assert event.app is mock_app
        assert event.shard is mock_shard
        assert event.integration is mock_app.entity_factory.deserialize_integration.return_value

    def test_deserialize_presence_update_event_with_only_user_id(self, event_factory, mock_app, mock_shard):
        mock_payload = {"user": {"id": "1231312"}}
        mock_old_presence = object()
        mock_app.entity_factory.deserialize_member_presence.return_value = mock.Mock(app=mock_app)

        event = event_factory.deserialize_presence_update_event(
            mock_shard, mock_payload, old_presence=mock_old_presence
        )

        mock_app.entity_factory.deserialize_member_presence.assert_called_once_with(mock_payload)
        assert isinstance(event, guild_events.PresenceUpdateEvent)
        assert event.shard is mock_shard
        assert event.old_presence is mock_old_presence
        assert event.user is None
        assert event.presence is mock_app.entity_factory.deserialize_member_presence.return_value

    def test_deserialize_presence_update_event_with_full_user_object(self, event_factory, mock_app, mock_shard):
        mock_payload = {
            "user": {
                "id": "1231312",
                "username": "OK",
                "global_name": "blahaj",
                "avatar": "NOK",
                "banner": "12122hssjamanmdd",
                "accent_color": 12342,
                "bot": True,
                "system": False,
                "public_flags": 42,
                "discriminator": "1231",
            }
        }
        mock_old_presence = mock.Mock(app=mock_app)

        event = event_factory.deserialize_presence_update_event(
            mock_shard, mock_payload, old_presence=mock_old_presence
        )

        mock_app.entity_factory.deserialize_member_presence.assert_called_once_with(mock_payload)
        assert isinstance(event, guild_events.PresenceUpdateEvent)
        assert event.shard is mock_shard
        assert event.old_presence is mock_old_presence

        assert isinstance(event.user, user_models.PartialUser)
        assert event.user.id == 1231312
        assert event.user.username == "OK"
        assert event.user.global_name == "blahaj"
        assert event.user.discriminator == "1231"
        assert event.user.avatar_hash == "NOK"
        assert event.user.banner_hash == "12122hssjamanmdd"
        assert event.user.accent_color == 12342
        assert event.user.is_bot is True
        assert event.user.is_system is False
        assert event.user.flags == 42

        assert event.presence is mock_app.entity_factory.deserialize_member_presence.return_value

    def test_deserialize_presence_update_event_with_partial_user_object(self, event_factory, mock_app, mock_shard):
        mock_payload = {"user": {"id": "1231312", "e": "OK"}}
        mock_old_presence = object()
        mock_app.entity_factory.deserialize_member_presence.return_value = mock.Mock(app=mock_app)

        event = event_factory.deserialize_presence_update_event(
            mock_shard, mock_payload, old_presence=mock_old_presence
        )

        mock_app.entity_factory.deserialize_member_presence.assert_called_once_with(mock_payload)
        assert isinstance(event, guild_events.PresenceUpdateEvent)
        assert event.shard is mock_shard
        assert event.old_presence is mock_old_presence

        assert isinstance(event.user, user_models.PartialUser)
        assert event.user.id == 1231312
        assert event.user.username is undefined.UNDEFINED
        assert event.user.global_name is undefined.UNDEFINED
        assert event.user.discriminator is undefined.UNDEFINED
        assert event.user.avatar_hash is undefined.UNDEFINED
        assert event.user.banner_hash is undefined.UNDEFINED
        assert event.user.accent_color is undefined.UNDEFINED
        assert event.user.is_bot is undefined.UNDEFINED
        assert event.user.is_system is undefined.UNDEFINED
        assert event.user.flags is undefined.UNDEFINED

        assert event.presence is mock_app.entity_factory.deserialize_member_presence.return_value

    def test_deserialize_audit_log_entry_create_event(
        self, event_factory: event_factory_.EventFactoryImpl, mock_app, mock_shard
    ):
        payload = {"id": "439034093490"}

        result = event_factory.deserialize_audit_log_entry_create_event(mock_shard, payload)

        mock_app.entity_factory.deserialize_audit_log_entry.assert_called_once_with(payload)
        assert result.entry is mock_app.entity_factory.deserialize_audit_log_entry.return_value
        assert result.shard is mock_shard
        assert isinstance(result, guild_events.AuditLogEntryCreateEvent)

    ######################
    # INTERACTION EVENTS #
    ######################

    @pytest.mark.parametrize(
        ("interaction_type", "expected"),
        [
            (base_interactions.InteractionType.APPLICATION_COMMAND, interaction_events.CommandInteractionCreateEvent),
            (base_interactions.InteractionType.MESSAGE_COMPONENT, interaction_events.ComponentInteractionCreateEvent),
            (base_interactions.InteractionType.AUTOCOMPLETE, interaction_events.AutocompleteInteractionCreateEvent),
            (base_interactions.InteractionType.MODAL_SUBMIT, interaction_events.ModalInteractionCreateEvent),
        ],
    )
    def test_deserialize_interaction_create_event(
        self,
        event_factory,
        mock_app,
        mock_shard,
        interaction_type: typing.Optional[base_interactions.InteractionType],
        expected: interaction_events.InteractionCreateEvent,
    ):
        payload = {"id": "1561232344"}
        if interaction_type:
            mock_app.entity_factory.deserialize_interaction.return_value = mock.Mock(type=interaction_type)
        result = event_factory.deserialize_interaction_create_event(mock_shard, payload)

        mock_app.entity_factory.deserialize_interaction.assert_called_once_with(payload)
        assert result.shard is mock_shard
        assert result.interaction is mock_app.entity_factory.deserialize_interaction.return_value
        assert isinstance(result, expected)

    def test_deserialize_interaction_create_event_error(self, event_factory, mock_app, mock_shard):
        payload = {"id": "1561232344"}
        with pytest.raises(KeyError):
            event_factory.deserialize_interaction_create_event(mock_shard, payload)

    #################
    # MEMBER EVENTS #
    #################

    def test_deserialize_guild_member_add_event(self, event_factory, mock_app, mock_shard):
        mock_payload = mock.Mock(app=mock_app)

        event = event_factory.deserialize_guild_member_add_event(mock_shard, mock_payload)

        mock_app.entity_factory.deserialize_member.assert_called_once_with(mock_payload)
        assert isinstance(event, member_events.MemberCreateEvent)
        assert event.shard is mock_shard
        assert event.member is mock_app.entity_factory.deserialize_member.return_value

    def test_deserialize_guild_member_update_event(self, event_factory, mock_app, mock_shard):
        mock_payload = mock.Mock(app=mock_app)
        mock_old_member = object()

        event = event_factory.deserialize_guild_member_update_event(
            mock_shard, mock_payload, old_member=mock_old_member
        )

        mock_app.entity_factory.deserialize_member.assert_called_once_with(mock_payload)
        assert isinstance(event, member_events.MemberUpdateEvent)
        assert event.shard is mock_shard
        assert event.member is mock_app.entity_factory.deserialize_member.return_value
        assert event.old_member is mock_old_member

    def test_deserialize_guild_member_remove_event(self, event_factory, mock_app, mock_shard):
        mock_user_payload = mock.Mock(app=mock_app)
        mock_old_member = object()
        mock_payload = {"guild_id": "43123", "user": mock_user_payload}

        event = event_factory.deserialize_guild_member_remove_event(
            mock_shard, mock_payload, old_member=mock_old_member
        )

        mock_app.entity_factory.deserialize_user.assert_called_once_with(mock_user_payload)
        assert isinstance(event, member_events.MemberDeleteEvent)
        assert event.shard is mock_shard
        assert event.guild_id == 43123
        assert event.user is mock_app.entity_factory.deserialize_user.return_value
        assert event.old_member is mock_old_member

    ###############
    # ROLE EVENTS #
    ###############

    def test_deserialize_guild_role_create_event(self, event_factory, mock_app, mock_shard):
        mock_role_payload = mock.Mock(app=mock_app)
        mock_payload = {"role": mock_role_payload, "guild_id": "45123"}

        event = event_factory.deserialize_guild_role_create_event(mock_shard, mock_payload)

        mock_app.entity_factory.deserialize_role.assert_called_once_with(mock_role_payload, guild_id=45123)
        assert isinstance(event, role_events.RoleCreateEvent)
        assert event.shard is mock_shard
        assert event.role is mock_app.entity_factory.deserialize_role.return_value

    def test_deserialize_guild_role_update_event(self, event_factory, mock_app, mock_shard):
        mock_role_payload = mock.Mock(app=mock_app)
        mock_old_role = object()
        mock_payload = {"role": mock_role_payload, "guild_id": "45123"}

        event = event_factory.deserialize_guild_role_update_event(mock_shard, mock_payload, old_role=mock_old_role)

        mock_app.entity_factory.deserialize_role.assert_called_once_with(mock_role_payload, guild_id=45123)
        assert isinstance(event, role_events.RoleUpdateEvent)
        assert event.shard is mock_shard
        assert event.role is mock_app.entity_factory.deserialize_role.return_value
        assert event.old_role is mock_old_role

    def test_deserialize_guild_role_delete_event(self, event_factory, mock_app, mock_shard):
        mock_payload = {"guild_id": "432123", "role_id": "848484"}
        mock_old_role = object()

        event = event_factory.deserialize_guild_role_delete_event(mock_shard, mock_payload, old_role=mock_old_role)

        assert isinstance(event, role_events.RoleDeleteEvent)
        assert event.app is mock_app
        assert event.shard is mock_shard
        assert event.guild_id == 432123
        assert event.role_id == 848484
        assert event.old_role is mock_old_role

    ##########################
    # SCHEDULED EVENT EVENTS #
    ##########################

    def test_deserialize_scheduled_event_create_event(
        self, event_factory: event_factory_.EventFactoryImpl, mock_app: traits.RESTAware, mock_shard: mock.Mock
    ):
        mock_payload = mock.Mock()

        event = event_factory.deserialize_scheduled_event_create_event(mock_shard, mock_payload)

        assert event.shard is mock_shard
        assert event.event is mock_app.entity_factory.deserialize_scheduled_event.return_value
        assert isinstance(event, scheduled_events.ScheduledEventCreateEvent)
        mock_app.entity_factory.deserialize_scheduled_event.assert_called_once_with(mock_payload)

    def test_deserialize_scheduled_event_update_event(
        self, event_factory: event_factory_.EventFactoryImpl, mock_app: traits.RESTAware, mock_shard: mock.Mock
    ):
        mock_payload = mock.Mock()

        event = event_factory.deserialize_scheduled_event_update_event(mock_shard, mock_payload)

        assert event.shard is mock_shard
        assert event.event is mock_app.entity_factory.deserialize_scheduled_event.return_value
        assert isinstance(event, scheduled_events.ScheduledEventUpdateEvent)
        mock_app.entity_factory.deserialize_scheduled_event.assert_called_once_with(mock_payload)

    def test_deserialize_scheduled_event_delete_event(
        self, event_factory: event_factory_.EventFactoryImpl, mock_app: traits.RESTAware, mock_shard: mock.Mock
    ):
        mock_payload = mock.Mock()

        event = event_factory.deserialize_scheduled_event_delete_event(mock_shard, mock_payload)

        assert event.shard is mock_shard
        assert event.event is mock_app.entity_factory.deserialize_scheduled_event.return_value
        assert isinstance(event, scheduled_events.ScheduledEventDeleteEvent)
        mock_app.entity_factory.deserialize_scheduled_event.assert_called_once_with(mock_payload)

    def test_deserialize_scheduled_event_user_add_event(
        self, event_factory: event_factory_.EventFactoryImpl, mock_app: mock.Mock, mock_shard: mock.Mock
    ):
        mock_payload = {"guild_id": "494949", "user_id": "123123123", "guild_scheduled_event_id": "49494944"}

        event = event_factory.deserialize_scheduled_event_user_add_event(mock_shard, mock_payload)

        assert event.shard is mock_shard
        assert event.guild_id == 494949
        assert event.user_id == 123123123
        assert event.event_id == 49494944
        assert isinstance(event, scheduled_events.ScheduledEventUserAddEvent)

    def test_deserialize_scheduled_event_user_remove_event(
        self, event_factory: event_factory_.EventFactoryImpl, mock_app: mock.Mock, mock_shard: mock.Mock
    ):
        mock_payload = {"guild_id": "3244321", "user_id": "56423", "guild_scheduled_event_id": "1234312"}

        event = event_factory.deserialize_scheduled_event_user_remove_event(mock_shard, mock_payload)

        assert event.shard is mock_shard
        assert event.guild_id == 3244321
        assert event.user_id == 56423
        assert event.event_id == 1234312
        assert isinstance(event, scheduled_events.ScheduledEventUserRemoveEvent)

    ###################
    # LIFETIME EVENTS #
    ###################

    def test_deserialize_starting_event(self, event_factory, mock_app):
        event = event_factory.deserialize_starting_event()

        assert isinstance(event, lifetime_events.StartingEvent)
        assert event.app is mock_app

    def test_deserialize_started_event(self, event_factory, mock_app):
        event = event_factory.deserialize_started_event()

        assert isinstance(event, lifetime_events.StartedEvent)
        assert event.app is mock_app

    def test_deserialize_stopping_event(self, event_factory, mock_app):
        event = event_factory.deserialize_stopping_event()

        assert isinstance(event, lifetime_events.StoppingEvent)
        assert event.app is mock_app

    def test_deserialize_stopped_event(self, event_factory, mock_app):
        event = event_factory.deserialize_stopped_event()

        assert isinstance(event, lifetime_events.StoppedEvent)
        assert event.app is mock_app

    ##################
    # MESSAGE EVENTS #
    ##################

    def test_deserialize_message_create_event_in_guild(self, event_factory, mock_app, mock_shard):
        mock_payload = mock.Mock(app=mock_app)
        mock_app.entity_factory.deserialize_message.return_value = mock.Mock(guild_id=123321)

        event = event_factory.deserialize_message_create_event(mock_shard, mock_payload)

        assert isinstance(event, message_events.GuildMessageCreateEvent)
        assert event.shard is mock_shard
        assert event.message is mock_app.entity_factory.deserialize_message.return_value

    def test_deserialize_message_create_event_in_dm(self, event_factory, mock_app, mock_shard):
        mock_payload = mock.Mock(app=mock_app)
        mock_app.entity_factory.deserialize_message.return_value = mock.Mock(guild_id=None)

        event = event_factory.deserialize_message_create_event(mock_shard, mock_payload)

        assert isinstance(event, message_events.DMMessageCreateEvent)
        assert event.shard is mock_shard
        assert event.message is mock_app.entity_factory.deserialize_message.return_value

    def test_deserialize_message_update_event_in_guild(self, event_factory, mock_app, mock_shard):
        mock_payload = mock.Mock(app=mock_app)
        mock_old_message = object()
        mock_app.entity_factory.deserialize_partial_message.return_value = mock.Mock(guild_id=123321, app=mock_app)

        event = event_factory.deserialize_message_update_event(mock_shard, mock_payload, old_message=mock_old_message)

        assert isinstance(event, message_events.GuildMessageUpdateEvent)
        assert event.shard is mock_shard
        assert event.message is mock_app.entity_factory.deserialize_partial_message.return_value
        assert event.old_message is mock_old_message

    def test_deserialize_message_update_event_in_dm(self, event_factory, mock_app, mock_shard):
        mock_payload = mock.Mock(app=mock_app)
        mock_old_message = object()
        mock_app.entity_factory.deserialize_partial_message.return_value = mock.Mock(guild_id=None)

        event = event_factory.deserialize_message_update_event(mock_shard, mock_payload, old_message=mock_old_message)

        assert isinstance(event, message_events.DMMessageUpdateEvent)
        assert event.shard is mock_shard
        assert event.message is mock_app.entity_factory.deserialize_partial_message.return_value
        assert event.old_message is mock_old_message

    def test_deserialize_message_delete_event_in_guild(self, event_factory, mock_app, mock_shard):
        mock_payload = {"id": "5412", "channel_id": "541123", "guild_id": "9494949"}
        old_message = object()

        event = event_factory.deserialize_message_delete_event(mock_shard, mock_payload, old_message=old_message)

        assert isinstance(event, message_events.GuildMessageDeleteEvent)
        assert event.app is mock_app
        assert event.shard is mock_shard
        assert event.old_message is old_message
        assert event.channel_id == 541123
        assert event.message_id == 5412
        assert event.guild_id == 9494949

    def test_deserialize_message_delete_event_in_dm(self, event_factory, mock_app, mock_shard):
        mock_payload = {"id": "5412", "channel_id": "541123"}
        old_message = object()

        event = event_factory.deserialize_message_delete_event(mock_shard, mock_payload, old_message=old_message)

        assert isinstance(event, message_events.DMMessageDeleteEvent)
        assert event.app is mock_app
        assert event.shard is mock_shard
        assert event.old_message is old_message
        assert event.channel_id == 541123
        assert event.message_id == 5412

    def test_deserialize_guild_message_delete_bulk_event(self, event_factory, mock_app, mock_shard):
        mock_payload = {"ids": ["6523423", "345123"], "channel_id": "564123", "guild_id": "4394949"}
        old_messages = object()

        event = event_factory.deserialize_guild_message_delete_bulk_event(
            mock_shard, mock_payload, old_messages=old_messages
        )

        assert isinstance(event, message_events.GuildBulkMessageDeleteEvent)
        assert event.app is mock_app
        assert event.shard is mock_shard
        assert event.old_messages is old_messages
        assert event.channel_id == 564123
        assert event.message_ids == {6523423, 345123}
        assert event.guild_id == 4394949

    def test_deserialize_guild_message_delete_bulk_event_when_old_messages_is_none(
        self, event_factory, mock_app, mock_shard
    ):
        mock_payload = {"ids": ["6523423", "345123"], "channel_id": "564123", "guild_id": "4394949"}

        event = event_factory.deserialize_guild_message_delete_bulk_event(mock_shard, mock_payload)

        assert isinstance(event, message_events.GuildBulkMessageDeleteEvent)
        assert event.old_messages == {}

    ###################
    # REACTION EVENTS #
    ###################

    def test_deserialize_message_reaction_add_event_in_guild(self, event_factory, mock_shard, mock_app):
        mock_member_payload = mock.Mock(app=mock_app)
        mock_payload = {
            "member": mock_member_payload,
            "channel_id": "34123",
            "message_id": "43123123",
            "guild_id": "43949494",
            "emoji": {"id": "123312", "name": "okok", "animated": True},
        }

        event = event_factory.deserialize_message_reaction_add_event(mock_shard, mock_payload)

        mock_app.entity_factory.deserialize_member.assert_called_once_with(mock_member_payload, guild_id=43949494)
        assert isinstance(event, reaction_events.GuildReactionAddEvent)
        assert event.shard is mock_shard
        assert event.channel_id == 34123
        assert event.message_id == 43123123
        assert event.member is mock_app.entity_factory.deserialize_member.return_value
        assert not isinstance(event.emoji_name, emoji_models.UnicodeEmoji)
        assert event.emoji_name == "okok"
        assert event.emoji_id == 123312
        assert event.is_animated is True

    def test_deserialize_message_reaction_add_event_in_guild_when_partial_custom(
        self, event_factory, mock_shard, mock_app
    ):
        mock_member_payload = object()
        mock_payload = {
            "member": mock_member_payload,
            "channel_id": "34123",
            "message_id": "43123123",
            "guild_id": "43949494",
            "emoji": {"id": "123312", "name": None},
        }

        event = event_factory.deserialize_message_reaction_add_event(mock_shard, mock_payload)

        assert event.is_animated is False
        assert event.emoji_id == 123312
        assert event.emoji_name is None

    def test_deserialize_message_reaction_add_event_in_guild_when_unicode(self, event_factory, mock_shard, mock_app):
        mock_member_payload = object()
        mock_payload = {
            "member": mock_member_payload,
            "channel_id": "34123",
            "message_id": "43123123",
            "guild_id": "43949494",
            "emoji": {"name": "hi", "id": None},
        }

        event = event_factory.deserialize_message_reaction_add_event(mock_shard, mock_payload)

        assert isinstance(event.emoji_name, emoji_models.UnicodeEmoji)
        assert event.emoji_name == "hi"
        assert event.emoji_id is None
        assert event.is_animated is False

    def test_deserialize_message_reaction_add_event_in_dm(self, event_factory, mock_shard, mock_app):
        mock_payload = {
            "channel_id": "34123",
            "message_id": "43123123",
            "user_id": "43949494",
            "emoji": {"id": "3293939", "name": "vohio", "animated": True},
        }

        event = event_factory.deserialize_message_reaction_add_event(mock_shard, mock_payload)

        assert isinstance(event, reaction_events.DMReactionAddEvent)
        assert event.app is mock_app
        assert event.shard is mock_shard
        assert event.channel_id == 34123
        assert event.message_id == 43123123
        assert event.user_id == 43949494
        assert not isinstance(event.emoji_name, emoji_models.UnicodeEmoji)
        assert event.emoji_name == "vohio"
        assert event.emoji_id == 3293939
        assert event.is_animated is True

    def test_deserialize_message_reaction_add_event_in_dm_when_partial_custom(
        self, event_factory, mock_shard, mock_app
    ):
        mock_payload = {
            "channel_id": "34123",
            "message_id": "43123123",
            "user_id": "43949494",
            "emoji": {"id": "3293939", "name": None},
        }

        event = event_factory.deserialize_message_reaction_add_event(mock_shard, mock_payload)

        assert event.emoji_name is None
        assert event.emoji_id == 3293939
        assert event.is_animated is False

    def test_deserialize_message_reaction_add_event_in_dm_when_unicode(self, event_factory, mock_shard, mock_app):
        mock_payload = {
            "channel_id": "34123",
            "message_id": "43123123",
            "user_id": "43949494",
            "emoji": {"name": "bye"},
        }

        event = event_factory.deserialize_message_reaction_add_event(mock_shard, mock_payload)

        assert isinstance(event, reaction_events.DMReactionAddEvent)
        assert event.app is mock_app
        assert event.shard is mock_shard
        assert event.channel_id == 34123
        assert event.message_id == 43123123
        assert event.user_id == 43949494
        assert isinstance(event.emoji_name, emoji_models.UnicodeEmoji)
        assert event.emoji_name == "bye"
        assert event.emoji_id is None
        assert event.is_animated is False

    def test_deserialize_message_reaction_remove_event_in_guild(self, event_factory, mock_app, mock_shard):
        mock_payload = {
            "user_id": "43123",
            "channel_id": "484848",
            "message_id": "43234",
            "guild_id": "383838",
            "emoji": {"id": "123432", "name": "fififiif"},
        }

        event = event_factory.deserialize_message_reaction_remove_event(mock_shard, mock_payload)

        assert isinstance(event, reaction_events.GuildReactionDeleteEvent)
        assert event.app is mock_app
        assert event.shard is mock_shard
        assert event.user_id == 43123
        assert event.channel_id == 484848
        assert event.message_id == 43234
        assert event.guild_id == 383838
        assert event.emoji_id == 123432
        assert event.emoji_name == "fififiif"
        assert not isinstance(event.emoji_name, emoji_models.UnicodeEmoji)

    def test_deserialize_message_reaction_remove_event_in_guild_with_unicode_emoji(
        self, event_factory, mock_app, mock_shard
    ):
        mock_payload = {
            "user_id": "43123",
            "channel_id": "484848",
            "message_id": "43234",
            "guild_id": "383838",
            "emoji": {"name": "o"},
        }

        event = event_factory.deserialize_message_reaction_remove_event(mock_shard, mock_payload)

        assert isinstance(event, reaction_events.GuildReactionDeleteEvent)
        assert event.app is mock_app
        assert event.shard is mock_shard
        assert event.user_id == 43123
        assert event.channel_id == 484848
        assert event.message_id == 43234
        assert event.guild_id == 383838
        assert event.emoji_id is None
        assert event.emoji_name == "o"
        assert isinstance(event.emoji_name, emoji_models.UnicodeEmoji)

    def test_deserialize_message_reaction_remove_event_in_dm(self, event_factory, mock_app, mock_shard):
        mock_payload = {
            "user_id": "43123",
            "channel_id": "484848",
            "message_id": "43234",
            "emoji": {"id": "123123", "name": "okok"},
        }

        event = event_factory.deserialize_message_reaction_remove_event(mock_shard, mock_payload)

        assert isinstance(event, reaction_events.DMReactionDeleteEvent)
        assert event.app is mock_app
        assert event.shard is mock_shard
        assert event.user_id == 43123
        assert event.channel_id == 484848
        assert event.message_id == 43234
        assert not isinstance(event.emoji_name, emoji_models.UnicodeEmoji)
        assert event.emoji_name == "okok"
        assert event.emoji_id == 123123

    def test_deserialize_message_reaction_remove_event_in_dm_with_unicode_emoji(
        self, event_factory, mock_app, mock_shard
    ):
        mock_payload = {"user_id": "43123", "channel_id": "484848", "message_id": "43234", "emoji": {"name": "wwww"}}

        event = event_factory.deserialize_message_reaction_remove_event(mock_shard, mock_payload)

        assert isinstance(event, reaction_events.DMReactionDeleteEvent)
        assert event.app is mock_app
        assert event.shard is mock_shard
        assert event.user_id == 43123
        assert event.channel_id == 484848
        assert event.message_id == 43234
        assert isinstance(event.emoji_name, emoji_models.UnicodeEmoji)
        assert event.emoji_name == "wwww"
        assert event.emoji_id is None

    def test_deserialize_message_reaction_remove_all_event_in_guild(self, event_factory, mock_app, mock_shard):
        mock_payload = {"channel_id": "312312", "message_id": "34323", "guild_id": "393939"}

        event = event_factory.deserialize_message_reaction_remove_all_event(mock_shard, mock_payload)

        assert isinstance(event, reaction_events.GuildReactionDeleteAllEvent)
        assert event.app is mock_app
        assert event.shard is mock_shard
        assert event.channel_id == 312312
        assert event.message_id == 34323
        assert event.guild_id == 393939

    def test_deserialize_message_reaction_remove_all_event_in_dm(self, event_factory, mock_app, mock_shard):
        mock_payload = {"channel_id": "312312", "message_id": "34323"}

        event = event_factory.deserialize_message_reaction_remove_all_event(mock_shard, mock_payload)

        assert isinstance(event, reaction_events.DMReactionDeleteAllEvent)
        assert event.app is mock_app
        assert event.shard is mock_shard
        assert event.channel_id == 312312
        assert event.message_id == 34323

    def test_deserialize_message_reaction_remove_emoji_event_in_guild(self, event_factory, mock_app, mock_shard):
        mock_payload = {
            "channel_id": "123123",
            "guild_id": "423412",
            "message_id": "99999",
            "emoji": {"id": "3123123", "name": "okokok"},
        }

        event = event_factory.deserialize_message_reaction_remove_emoji_event(mock_shard, mock_payload)

        assert isinstance(event, reaction_events.GuildReactionDeleteEmojiEvent)
        assert event.app is mock_app
        assert event.shard is mock_shard
        assert event.channel_id == 123123
        assert event.guild_id == 423412
        assert event.message_id == 99999
        assert event.emoji_id == 3123123
        assert event.emoji_name == "okokok"
        assert not isinstance(event.emoji_name, emoji_models.UnicodeEmoji)

    def test_deserialize_message_reaction_remove_emoji_event_in_guild_with_unicode_emoji(
        self, event_factory, mock_app, mock_shard
    ):
        mock_payload = {
            "channel_id": "123123",
            "guild_id": "423412",
            "message_id": "99999",
            "emoji": {"name": "okokok"},
        }

        event = event_factory.deserialize_message_reaction_remove_emoji_event(mock_shard, mock_payload)

        assert event.emoji_name == "okokok"
        assert isinstance(event.emoji_name, emoji_models.UnicodeEmoji)

    def test_deserialize_message_reaction_remove_emoji_event_in_dm(self, event_factory, mock_app, mock_shard):
        mock_payload = {"channel_id": "123123", "message_id": "99999", "emoji": {"id": "123321", "name": "nom"}}

        event = event_factory.deserialize_message_reaction_remove_emoji_event(mock_shard, mock_payload)

        assert isinstance(event, reaction_events.DMReactionDeleteEmojiEvent)
        assert event.app is mock_app
        assert event.shard is mock_shard
        assert event.channel_id == 123123
        assert event.message_id == 99999
        assert event.emoji_id == 123321
        assert event.emoji_name == "nom"
        assert not isinstance(event.emoji_name, emoji_models.UnicodeEmoji)

    def test_deserialize_message_reaction_remove_emoji_event_in_dm_with_unicode_emoji(
        self, event_factory, mock_app, mock_shard
    ):
        mock_payload = {"channel_id": "123123", "message_id": "99999", "emoji": {"name": "gg"}}

        event = event_factory.deserialize_message_reaction_remove_emoji_event(mock_shard, mock_payload)

        assert isinstance(event, reaction_events.DMReactionDeleteEmojiEvent)
        assert event.app is mock_app
        assert event.shard is mock_shard
        assert event.channel_id == 123123
        assert event.message_id == 99999
        assert event.emoji_id is None
        assert event.emoji_name == "gg"
        assert isinstance(event.emoji_name, emoji_models.UnicodeEmoji)

    ################
    # SHARD EVENTS #
    ################

    def test_deserialize_shard_payload_event(self, event_factory, mock_app, mock_shard):
        mock_payload = {"id": "123123"}

        event = event_factory.deserialize_shard_payload_event(mock_shard, mock_payload, name="ooga booga")

        assert event.app is mock_app
        assert event.name == "ooga booga"
        assert event.payload == mock_payload
        assert event.shard is mock_shard

    def test_deserialize_ready_event(self, event_factory, mock_app, mock_shard):
        mock_user_payload = object()
        mock_payload = {
            "v": "69",
            "resume_gateway_url": "testing.com",
            "user": mock_user_payload,
            "guilds": [{"id": "432123"}, {"id": "949494"}],
            "session_id": "kjsdjiodsaiosad",
            "application": {"id": "4123212", "flags": "4949494"},
        }
        mock_app.entity_factory.deserialize_my_user.return_value = mock.Mock(app=mock_app)

        event = event_factory.deserialize_ready_event(mock_shard, mock_payload)

        mock_app.entity_factory.deserialize_my_user.assert_called_once_with(mock_user_payload)
        assert isinstance(event, shard_events.ShardReadyEvent)
        assert event.shard is mock_shard
        assert event.actual_gateway_version == 69
        assert event.resume_gateway_url == "testing.com"
        assert event.my_user is mock_app.entity_factory.deserialize_my_user.return_value
        assert event.unavailable_guilds == [432123, 949494]
        assert event.session_id == "kjsdjiodsaiosad"
        assert event.application_id == 4123212
        assert event.application_flags == 4949494

    def test_deserialize_connected_event(self, event_factory, mock_app, mock_shard):
        event = event_factory.deserialize_connected_event(mock_shard)

        assert isinstance(event, shard_events.ShardConnectedEvent)
        assert event.app is mock_app
        assert event.shard is mock_shard

    def test_deserialize_disconnected_event(self, event_factory, mock_app, mock_shard):
        event = event_factory.deserialize_disconnected_event(mock_shard)

        assert isinstance(event, shard_events.ShardDisconnectedEvent)
        assert event.app is mock_app
        assert event.shard is mock_shard

    def test_deserialize_resumed_event(self, event_factory, mock_app, mock_shard):
        event = event_factory.deserialize_resumed_event(mock_shard)

        assert isinstance(event, shard_events.ShardResumedEvent)
        assert event.app is mock_app
        assert event.shard is mock_shard

    def test_deserialize_guild_member_chunk_event_with_optional_fields(self, event_factory, mock_app, mock_shard):
        mock_member_payload = {"user": {"id": "4222222"}}
        mock_presence_payload = {"user": {"id": "43123123"}}
        mock_payload = {
            "guild_id": "123432123",
            "members": [mock_member_payload],
            "chunk_index": 3,
            "chunk_count": 54,
            "not_found": ["34212312312", 323123123],
            "presences": [mock_presence_payload],
            "nonce": "OKOKOKOK",
        }

        event = event_factory.deserialize_guild_member_chunk_event(mock_shard, mock_payload)

        mock_app.entity_factory.deserialize_member.assert_called_once_with(mock_member_payload, guild_id=123432123)
        mock_app.entity_factory.deserialize_member_presence.assert_called_once_with(
            mock_presence_payload, guild_id=123432123
        )
        assert isinstance(event, shard_events.MemberChunkEvent)
        assert event.app is mock_app
        assert event.shard is mock_shard
        assert event.guild_id == 123432123
        assert event.members == {4222222: mock_app.entity_factory.deserialize_member.return_value}
        assert event.chunk_count == 54
        assert event.chunk_index == 3
        assert event.not_found == [34212312312, 323123123]
        assert event.presences == {43123123: mock_app.entity_factory.deserialize_member_presence.return_value}
        assert event.nonce == "OKOKOKOK"

    def test_deserialize_guild_member_chunk_event_without_optional_fields(self, event_factory, mock_app, mock_shard):
        mock_member_payload = {"user": {"id": "4222222"}}
        mock_payload = {"guild_id": "123432123", "members": [mock_member_payload], "chunk_index": 3, "chunk_count": 54}

        event = event_factory.deserialize_guild_member_chunk_event(mock_shard, mock_payload)

        assert event.not_found == []
        assert event.presences == {}
        assert event.nonce is None

    ###############
    # USER EVENTS #
    ###############

    def test_deserialize_own_user_update_event(self, event_factory, mock_app, mock_shard):
        mock_payload = mock.Mock(app=mock_app)
        mock_old_user = object()
        mock_app.entity_factory.deserialize_my_user.return_value = mock.Mock(app=mock_app)

        event = event_factory.deserialize_own_user_update_event(mock_shard, mock_payload, old_user=mock_old_user)

        mock_app.entity_factory.deserialize_my_user.assert_called_once_with(mock_payload)
        assert isinstance(event, user_events.OwnUserUpdateEvent)
        assert event.shard is mock_shard
        assert event.user is mock_app.entity_factory.deserialize_my_user.return_value
        assert event.old_user is mock_old_user

    ################
    # VOICE EVENTS #
    ################

    def test_deserialize_voice_state_update_event(self, event_factory, mock_app, mock_shard):
        mock_payload = object()
        mock_old_voice_state = object()
        mock_app.entity_factory.deserialize_voice_state.return_value = mock.Mock(app=mock_app)

        event = event_factory.deserialize_voice_state_update_event(
            mock_shard, mock_payload, old_state=mock_old_voice_state
        )

        mock_app.entity_factory.deserialize_voice_state.assert_called_once_with(mock_payload)
        assert isinstance(event, voice_events.VoiceStateUpdateEvent)
        assert event.shard is mock_shard
        assert event.state is mock_app.entity_factory.deserialize_voice_state.return_value
        assert event.old_state is mock_old_voice_state

    def test_deserialize_voice_server_update_event(self, event_factory, mock_app, mock_shard):
        mock_payload = {"token": "okokok", "guild_id": "3122312", "endpoint": "httppppppp"}

        event = event_factory.deserialize_voice_server_update_event(mock_shard, mock_payload)

        assert isinstance(event, voice_events.VoiceServerUpdateEvent)
        assert event.app is mock_app
        assert event.shard is mock_shard
        assert event.token == "okokok"
        assert event.guild_id == 3122312
        assert event.raw_endpoint == "httppppppp"

    ##################
    #  MONETIZATION  #
    ##################

    def test_deserialize_entitlement_create_event(self, event_factory, mock_app, mock_shard):
        payload = {
            "id": "696969696969696",
            "sku_id": "420420420420420",
            "application_id": "123123123123123",
            "type": 8,
            "deleted": False,
            "starts_at": "2022-09-14T17:00:18.704163+00:00",
            "ends_at": "2022-10-14T17:00:18.704163+00:00",
            "guild_id": "1015034326372454400",
            "user_id": "115590097100865541",
            "subscription_id": "1019653835926409216",
        }

        event = event_factory.deserialize_entitlement_create_event(mock_shard, payload)

        assert isinstance(event, monetization_events.EntitlementCreateEvent)

    def test_deserialize_entitlement_update_event(self, event_factory, mock_app, mock_shard):
        payload = {
            "id": "696969696969696",
            "sku_id": "420420420420420",
            "application_id": "123123123123123",
            "type": 8,
            "deleted": False,
            "starts_at": "2022-09-14T17:00:18.704163+00:00",
            "ends_at": "2022-10-14T17:00:18.704163+00:00",
            "guild_id": "1015034326372454400",
            "user_id": "115590097100865541",
            "subscription_id": "1019653835926409216",
        }

        event = event_factory.deserialize_entitlement_update_event(mock_shard, payload)

        assert isinstance(event, monetization_events.EntitlementUpdateEvent)

    def test_deserialize_entitlement_delete_event(self, event_factory, mock_app, mock_shard):
        payload = {
            "id": "696969696969696",
            "sku_id": "420420420420420",
            "application_id": "123123123123123",
            "type": 8,
            "deleted": False,
            "starts_at": "2022-09-14T17:00:18.704163+00:00",
            "ends_at": "2022-10-14T17:00:18.704163+00:00",
            "guild_id": "1015034326372454400",
            "user_id": "115590097100865541",
            "subscription_id": "1019653835926409216",
        }

        event = event_factory.deserialize_entitlement_delete_event(mock_shard, payload)

        assert isinstance(event, monetization_events.EntitlementDeleteEvent)

    #########################
    # STAGE INSTANCE EVENTS #
    #########################

    def test_deserialize_stage_instance_create_event(self, event_factory, mock_app, mock_shard):
        mock_payload = {
            "id": "840647391636226060",
            "guild_id": "197038439483310086",
            "channel_id": "733488538393510049",
            "topic": "Testing Testing, 123",
            "privacy_level": 1,
            "discoverable_disabled": False,
        }
        event = event_factory.deserialize_stage_instance_create_event(mock_shard, mock_payload)
        assert isinstance(event, stage_events.StageInstanceCreateEvent)

        assert event.shard is mock_shard
        assert event.app is event.stage_instance.app
        assert event.stage_instance == mock_app.entity_factory.deserialize_stage_instance.return_value

    def test_deserialize_stage_instance_update_event(self, event_factory, mock_app, mock_shard):
        mock_payload = {
            "id": "840647391636226060",
            "guild_id": "197038439483310086",
            "channel_id": "733488538393510049",
            "topic": "Testing Testing, 124",
            "privacy_level": 2,
            "discoverable_disabled": True,
        }
        event = event_factory.deserialize_stage_instance_update_event(mock_shard, mock_payload)
        assert isinstance(event, stage_events.StageInstanceUpdateEvent)

        assert event.shard is mock_shard
        assert event.app is event.stage_instance.app
        assert event.stage_instance == mock_app.entity_factory.deserialize_stage_instance.return_value

    def test_deserialize_stage_instance_delete_event(self, event_factory, mock_app, mock_shard):
        mock_payload = {
            "id": "840647391636226060",
            "guild_id": "197038439483310086",
            "channel_id": "733488538393510049",
            "topic": "Testing Testing, 124",
            "privacy_level": 2,
            "discoverable_disabled": True,
        }
        event = event_factory.deserialize_stage_instance_delete_event(mock_shard, mock_payload)
        assert isinstance(event, stage_events.StageInstanceDeleteEvent)

        assert event.shard is mock_shard
        assert event.app is event.stage_instance.app
        assert event.stage_instance == mock_app.entity_factory.deserialize_stage_instance.return_value

    ###########
    #  POLLS  #
    ###########

    def test_deserialize_poll_vote_create_event(self, event_factory, mock_app, mock_shard):
        payload = {
            "user_id": "3847382",
            "channel_id": "4598743",
            "message_id": "458437954",
            "guild_id": "3589273",
            "answer_id": 1,
        }

        event = event_factory.deserialize_poll_vote_create_event(mock_shard, payload)

        assert isinstance(event, poll_events.PollVoteCreateEvent)

    def test_deserialize_poll_vote_delete_event(self, event_factory, mock_app, mock_shard):
        payload = {
            "user_id": "3847382",
            "channel_id": "4598743",
            "message_id": "458437954",
            "guild_id": "3589273",
            "answer_id": 1,
        }

        event = event_factory.deserialize_poll_vote_delete_event(mock_shard, payload)

        assert isinstance(event, poll_events.PollVoteDeleteEvent)

    def test_deserialize_auto_mod_rule_create_event(self, event_factory, mock_app, mock_shard):
        mock_payload = {"id": "49499494"}

        event = event_factory.deserialize_auto_mod_rule_create_event(mock_shard, mock_payload)

        assert isinstance(event, auto_mod_events.AutoModRuleCreateEvent)
        assert event.shard is mock_shard
        assert event.rule is mock_app.entity_factory.deserialize_auto_mod_rule.return_value
        mock_app.entity_factory.deserialize_auto_mod_rule.assert_called_once_with(mock_payload)

    def test_deserialize_auto_mod_rule_update_event(self, event_factory, mock_app, mock_shard):
        mock_payload = {"id": "49499494"}

        event = event_factory.deserialize_auto_mod_rule_update_event(mock_shard, mock_payload)

        assert isinstance(event, auto_mod_events.AutoModRuleUpdateEvent)
        assert event.shard is mock_shard
        assert event.rule is mock_app.entity_factory.deserialize_auto_mod_rule.return_value
        mock_app.entity_factory.deserialize_auto_mod_rule.assert_called_once_with(mock_payload)

    def test_deserialize_auto_mod_rule_delete_event(self, event_factory, mock_app, mock_shard):
        mock_payload = {"id": "49499494"}

        event = event_factory.deserialize_auto_mod_rule_delete_event(mock_shard, mock_payload)

        assert isinstance(event, auto_mod_events.AutoModRuleDeleteEvent)
        assert event.shard is mock_shard
        assert event.rule is mock_app.entity_factory.deserialize_auto_mod_rule.return_value
        mock_app.entity_factory.deserialize_auto_mod_rule.assert_called_once_with(mock_payload)

    def test_deserialize_auto_mod_action_execution_event(self, event_factory, mock_app, mock_shard):
        mock_action_payload = {"type": "69"}

        event = event_factory.deserialize_auto_mod_action_execution_event(
            mock_shard,
            {
                "guild_id": "123321",
                "action": mock_action_payload,
                "rule_id": "4959595",
                "rule_trigger_type": 3,
                "user_id": "4949494",
                "channel_id": "5423234",
                "message_id": "49343292",
                "alert_system_message_id": "49211123",
                "content": "meow",
                "matched_keyword": "fredf",
                "matched_content": "dfofodofdodf",
            },
        )

        assert event.app is mock_app
        assert event.shard is mock_shard
        assert event.guild_id == 123321
        assert event.action is mock_app.entity_factory.deserialize_auto_mod_action.return_value
        assert event.rule_id == 4959595
        assert event.rule_trigger_type is auto_mod.AutoModTriggerType.SPAM
        assert event.user_id == 4949494
        assert event.channel_id == 5423234
        assert event.message_id == 49343292
        assert event.alert_system_message_id == 49211123
        assert event.content == "meow"
        assert event.matched_keyword == "fredf"
        assert event.matched_content == "dfofodofdodf"
        mock_app.entity_factory.deserialize_auto_mod_action.assert_called_once_with(mock_action_payload)

    def test_deserialize_auto_mod_action_execution_event_when_partial(self, event_factory, mock_app, mock_shard):
        mock_action_payload = {"type": "69"}

        event = event_factory.deserialize_auto_mod_action_execution_event(
            mock_shard,
            {
                "guild_id": "123321",
                "action": mock_action_payload,
                "rule_id": "4959595",
                "rule_trigger_type": 3,
                "user_id": "4949494",
                "content": "",
                "matched_keyword": None,
                "matched_content": None,
            },
        )

        assert event.app is mock_app
        assert event.shard is mock_shard
        assert event.guild_id == 123321
        assert event.action is mock_app.entity_factory.deserialize_auto_mod_action.return_value
        assert event.rule_id == 4959595
        assert event.rule_trigger_type is auto_mod.AutoModTriggerType.SPAM
        assert event.user_id == 4949494
        assert event.channel_id is None
        assert event.message_id is None
        assert event.alert_system_message_id is None
        assert event.content is None
        assert event.matched_keyword is None
        assert event.matched_content is None
        mock_app.entity_factory.deserialize_auto_mod_action.assert_called_once_with(mock_action_payload)



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Bitfield of permissions."""

from __future__ import annotations

__all__: typing.Sequence[str] = ("Permissions",)

import typing

from hikari.internal import enums


@typing.final
class Permissions(enums.Flag):
    """Represents the permissions available in a given channel or guild.

    This enum is an [`enum.IntFlag`][], which means that it is stored as a bit field
    where each bit represents a permission. You can use bitwise operators
    to efficiently manipulate and compare permissions.

    Examples
    --------
    You can create an enum which combines multiple permissions using the bitwise OR operator (`|`):

    ```py
    my_perms = Permissions.MANAGE_CHANNELS | Permissions.MANAGE_GUILD

    required_perms = (
        Permissions.CREATE_INSTANT_INVITE
        | Permissions.KICK_MEMBERS
        | Permissions.BAN_MEMBERS
        | Permissions.MANAGE_GUILD
    )
    ```

    To find the intersection of two sets of permissions, use the bitwise AND
    operator (`&`) between them. By then applying the `==` operator, you can check if all
    permissions from one set are present in another set. This is useful, for instance,
    for checking if a user has all the required permissions

    ```py
    if (my_perms & required_perms) == required_perms:
        print("I have all of the required permissions!")
    else:
        print("I am missing at least one required permission!")
    ```

    To determine which permissions from one set are missing from another, you can use the
    bitwise equivalent of the set difference operation, as shown below. This can be used,
    for instance, to find which of a user's permissions are missing from the required permissions.

    ```py
    missing_perms = ~my_perms & required_perms
    if missing_perms:
        print(f"I'm missing these permissions: {missing_perms}")
    ```

    Lastly, if you need all the permissions from a set except for a few,
    you can use the bitwise NOT operator (`~`).

    ```py
    # All permissions except ADMINISTRATOR.
    my_perms = ~Permissions.ADMINISTRATOR
    ```
    """

    NONE = 0
    """Empty permission."""

    CREATE_INSTANT_INVITE = 1 << 0
    """Allows creation of instant invites."""

    KICK_MEMBERS = 1 << 1
    """Allows kicking members.

    !!! note
        In guilds with server-wide 2FA enabled this permission can only be used
        by users who have two-factor authentication enabled on their account
        (or their owner's account in the case of bot users) and the guild owner.
    """

    BAN_MEMBERS = 1 << 2
    """Allows banning members.

    !!! note
        In guilds with server-wide 2FA enabled this permission can only be used
        by users who have two-factor authentication enabled on their account
        (or their owner's account in the case of bot users) and the guild owner.
    """

    ADMINISTRATOR = 1 << 3
    """Allows all permissions and bypasses channel permission overwrites.

    !!! note
        In guilds with server-wide 2FA enabled this permission can only be used
        by users who have two-factor authentication enabled on their account
        (or their owner's account in the case of bot users) and the guild owner.
    """

    MANAGE_CHANNELS = 1 << 4
    """Allows management and editing of channels.

    !!! note
        In guilds with server-wide 2FA enabled this permission can only be used
        by users who have two-factor authentication enabled on their account
        (or their owner's account in the case of bot users) and the guild owner.
    """

    MANAGE_GUILD = 1 << 5
    """Allows management and editing of the guild.

    !!! note
        In guilds with server-wide 2FA enabled this permission can only be used
        by users who have two-factor authentication enabled on their account
        (or their owner's account in the case of bot users) and the guild owner.
    """

    ADD_REACTIONS = 1 << 6
    """Allows for the addition of reactions to messages."""

    VIEW_AUDIT_LOG = 1 << 7
    """Allows for viewing of audit logs."""

    PRIORITY_SPEAKER = 1 << 8
    """Allows for using priority speaker in a voice channel."""

    STREAM = 1 << 9
    """Allows the user to go live."""

    VIEW_CHANNEL = 1 << 10
    """Allows guild members to view a channel, which includes reading messages in text channels."""

    SEND_MESSAGES = 1 << 11
    """Allows for sending messages in a channel."""

    SEND_TTS_MESSAGES = 1 << 12
    """Allows for sending of Text-To-Speech (`/tts`) messages."""

    MANAGE_MESSAGES = 1 << 13
    """Allows for deletion of other users messages.

    !!! note
        In guilds with server-wide 2FA enabled this permission can only be used
        by users who have two-factor authentication enabled on their account
        (or their owner's account in the case of bot users) and the guild owner.
    """

    EMBED_LINKS = 1 << 14
    """Links sent by users with this permission will be auto-embedded."""

    ATTACH_FILES = 1 << 15
    """Allows for uploading images and files."""

    READ_MESSAGE_HISTORY = 1 << 16
    """Allows for reading of message history."""

    MENTION_ROLES = 1 << 17
    """Allows for using the `@everyone`, `@here` and `@role` (regardless of its mention status) tag to notify users."""

    USE_EXTERNAL_EMOJIS = 1 << 18
    """Allows the usage of custom emojis from other guilds."""

    VIEW_GUILD_INSIGHTS = 1 << 19
    """Allows the user to view guild insights for eligible guilds."""

    CONNECT = 1 << 20
    """Allows for joining of a voice channel."""

    SPEAK = 1 << 21
    """Allows for speaking in a voice channel."""

    MUTE_MEMBERS = 1 << 22
    """Allows for muting members in a voice channel."""

    DEAFEN_MEMBERS = 1 << 23
    """Allows for deafening of members in a voice channel."""

    MOVE_MEMBERS = 1 << 24
    """Allows for moving of members between voice channels."""

    USE_VOICE_ACTIVITY = 1 << 25
    """Allows for using voice-activity-detection in a voice channel."""

    CHANGE_NICKNAME = 1 << 26
    """Allows for modification of own nickname."""

    MANAGE_NICKNAMES = 1 << 27
    """Allows for modification of other users nicknames."""

    MANAGE_ROLES = 1 << 28
    """Allows management and editing of roles.

    !!! note
        In guilds with server-wide 2FA enabled this permission can only be used
        by users who have two-factor authentication enabled on their account
        (or their owner's account in the case of bot users) and the guild owner.
    """

    MANAGE_WEBHOOKS = 1 << 29
    """Allows management and editing of webhooks.

    !!! note
        In guilds with server-wide 2FA enabled this permission can only be used
        by users who have two-factor authentication enabled on their account
        (or their owner's account in the case of bot users) and the guild owner.
    """

    MANAGE_GUILD_EXPRESSIONS = 1 << 30
    """Allows management and editing emojis, stickers and soundboard sounds.

    !!! note
        In guilds with server-wide 2FA enabled this permission can only be used
        by users who have two-factor authentication enabled on their account
        (or their owner's account in the case of bot users) and the guild owner.
    """

    USE_APPLICATION_COMMANDS = 1 << 31
    """Allows for using the application commands of guild integrations within a text channel."""

    REQUEST_TO_SPEAK = 1 << 32
    """Allows for requesting to speak in stage channels.

    !!! warning
        This permissions is currently defined as being "under active
        development" by Discord meaning that "it may be changed or removed"
        without warning.
    """

    MANAGE_EVENTS = 1 << 33
    """Allows for management and editing scheduled events"""

    MANAGE_THREADS = 1 << 34
    """Allows for deleting and archiving threads, and viewing all private threads.

    !!! note
        In guilds with server-wide 2FA enabled this permission can only be used
        by users who have two-factor authentication enabled on their account
        (or their owner's account in the case of bot users) and the guild owner.
    """

    CREATE_PUBLIC_THREADS = 1 << 35
    """Allows for creating threads."""

    CREATE_PRIVATE_THREADS = 1 << 36
    """Allows for creating private threads."""

    USE_EXTERNAL_STICKERS = 1 << 37
    """Allows the usage of custom stickers from other servers."""

    SEND_MESSAGES_IN_THREADS = 1 << 38
    """Allows for sending messages in threads."""

    START_EMBEDDED_ACTIVITIES = 1 << 39
    """Allows for launching activities in a voice channel.

    Activities are applications that have the [`hikari.applications.ApplicationFlags.EMBEDDED`][] flag.
    """

    MODERATE_MEMBERS = 1 << 40
    """Allows for timing out members."""

    VIEW_CREATOR_MONETIZATION_ANALYTICS = 1 << 41
    """Allows for viewing role subscription insights"""

    USE_SOUNDBOARD = 1 << 42
    """Allows the use of soundboard in a voice chat."""

    CREATE_GUILD_EXPRESSIONS = 1 << 43
    """Allows to create guild emojis, stickers and soundboard sounds.

    Additionally, it allows to edit and manage those created by the user.
    """

    CREATE_EVENTS = 1 << 44
    """Allows to create scheduled events.

    Additionally, it allows to edit and manage those created by the user.
    """

    USE_EXTERNAL_SOUNDS = 1 << 45
    """Allows the use of soundboard sounds from external servers."""

    SEND_VOICE_MESSAGES = 1 << 46
    """Allows sending voice messages."""

    SEND_POLLS = 1 << 49
    """Allows sending polls."""

    USE_EXTERNAL_APPS = 1 << 50
    """Allows user-installed apps to send public responses.

    When disabled, users will still be allowed to use their apps but the responses will be ephemeral.

    This only applies to apps not also installed to the server.
    """

    @classmethod
    def all_permissions(cls) -> Permissions:
        """Get an instance of [`hikari.permissions.Permissions`][] with all the known permissions.

        Returns
        -------
        Permissions
            A permissions instance with all the known permissions.
        """
        all_perms = Permissions.NONE
        for perm in Permissions:
            all_perms |= perm

        return all_perms



# DO NOT MANUALLY EDIT THIS FILE!
# This file was automatically generated by `nox -s generate-stubs`

from hikari import api as api
from hikari import applications as applications
from hikari import events as events
from hikari import files as files
from hikari import impl as impl
from hikari import interactions as interactions
from hikari import snowflakes as snowflakes
from hikari import undefined as undefined
from hikari._about import __author__ as __author__
from hikari._about import __ci__ as __ci__
from hikari._about import __copyright__ as __copyright__
from hikari._about import __coverage__ as __coverage__
from hikari._about import __discord_invite__ as __discord_invite__
from hikari._about import __docs__ as __docs__
from hikari._about import __email__ as __email__
from hikari._about import __git_sha1__ as __git_sha1__
from hikari._about import __issue_tracker__ as __issue_tracker__
from hikari._about import __license__ as __license__
from hikari._about import __maintainer__ as __maintainer__
from hikari._about import __url__ as __url__
from hikari._about import __version__ as __version__
from hikari.applications import Application as Application
from hikari.applications import ApplicationContextType as ApplicationContextType
from hikari.applications import ApplicationFlags as ApplicationFlags
from hikari.applications import ApplicationIntegrationType as ApplicationIntegrationType
from hikari.applications import ApplicationRoleConnectionMetadataRecord as ApplicationRoleConnectionMetadataRecord
from hikari.applications import (
    ApplicationRoleConnectionMetadataRecordType as ApplicationRoleConnectionMetadataRecordType,
)
from hikari.applications import AuthorizationApplication as AuthorizationApplication
from hikari.applications import AuthorizationInformation as AuthorizationInformation
from hikari.applications import ConnectionVisibility as ConnectionVisibility
from hikari.applications import OAuth2AuthorizationToken as OAuth2AuthorizationToken
from hikari.applications import OAuth2ImplicitToken as OAuth2ImplicitToken
from hikari.applications import OAuth2Scope as OAuth2Scope
from hikari.applications import OwnApplicationRoleConnection as OwnApplicationRoleConnection
from hikari.applications import OwnConnection as OwnConnection
from hikari.applications import OwnGuild as OwnGuild
from hikari.applications import PartialOAuth2Token as PartialOAuth2Token
from hikari.applications import Team as Team
from hikari.applications import TeamMember as TeamMember
from hikari.applications import TeamMembershipState as TeamMembershipState
from hikari.applications import TokenType as TokenType
from hikari.audit_logs import *
from hikari.auto_mod import *
from hikari.channels import *
from hikari.colors import *
from hikari.colours import *
from hikari.commands import *
from hikari.components import *
from hikari.embeds import *
from hikari.emojis import *
from hikari.errors import *
from hikari.events.base_events import Event as Event
from hikari.events.base_events import ExceptionEvent as ExceptionEvent
from hikari.events.channel_events import *
from hikari.events.guild_events import *
from hikari.events.interaction_events import *
from hikari.events.lifetime_events import *
from hikari.events.member_events import *
from hikari.events.message_events import *
from hikari.events.monetization_events import *
from hikari.events.poll_events import *
from hikari.events.reaction_events import *
from hikari.events.role_events import *
from hikari.events.scheduled_events import *
from hikari.events.shard_events import *
from hikari.events.typing_events import *
from hikari.events.user_events import *
from hikari.events.voice_events import *
from hikari.files import URL as URL
from hikari.files import Bytes as Bytes
from hikari.files import File as File
from hikari.files import LazyByteIteratorish as LazyByteIteratorish
from hikari.files import Pathish as Pathish
from hikari.files import Rawish as Rawish
from hikari.files import Resourceish as Resourceish
from hikari.guilds import *
from hikari.impl import ClientCredentialsStrategy as ClientCredentialsStrategy
from hikari.impl import GatewayBot as GatewayBot
from hikari.impl import RESTApp as RESTApp
from hikari.impl import RESTBot as RESTBot
from hikari.intents import *
from hikari.interactions.base_interactions import *
from hikari.interactions.command_interactions import *
from hikari.interactions.component_interactions import *
from hikari.interactions.modal_interactions import *
from hikari.invites import *
from hikari.iterators import *
from hikari.locales import *
from hikari.messages import *
from hikari.monetization import *
from hikari.permissions import *
from hikari.polls import *
from hikari.presences import *
from hikari.scheduled_events import *
from hikari.sessions import *
from hikari.snowflakes import SearchableSnowflakeish as SearchableSnowflakeish
from hikari.snowflakes import SearchableSnowflakeishOr as SearchableSnowflakeishOr
from hikari.snowflakes import Snowflake as Snowflake
from hikari.snowflakes import Snowflakeish as Snowflakeish
from hikari.snowflakes import SnowflakeishOr as SnowflakeishOr
from hikari.snowflakes import SnowflakeishSequence as SnowflakeishSequence
from hikari.snowflakes import Unique as Unique
from hikari.stage_instances import *
from hikari.stickers import *
from hikari.templates import *
from hikari.traits import *
from hikari.undefined import UNDEFINED as UNDEFINED
from hikari.undefined import UndefinedNoneOr as UndefinedNoneOr
from hikari.undefined import UndefinedOr as UndefinedOr
from hikari.undefined import UndefinedType as UndefinedType
from hikari.users import *
from hikari.voices import *
from hikari.webhooks import *



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import http

import aiohttp
import mock
import pytest

from hikari import errors
from hikari.internal import net


@pytest.mark.parametrize(
    ("status_", "expected_error"),
    [
        (http.HTTPStatus.UNAUTHORIZED, "UnauthorizedError"),
        (http.HTTPStatus.FORBIDDEN, "ForbiddenError"),
        (http.HTTPStatus.NOT_FOUND, "NotFoundError"),
        (http.HTTPStatus.PAYMENT_REQUIRED, "ClientHTTPResponseError"),
        (http.HTTPStatus.SERVICE_UNAVAILABLE, "InternalServerError"),
        (http.HTTPStatus.PERMANENT_REDIRECT, "HTTPResponseError"),
    ],
)
@pytest.mark.asyncio
async def test_generate_error_response(status_, expected_error):
    class StubResponse:
        real_url = "https://some.url"
        status = status_
        headers = {}

        async def read(self):
            return '{"message": "raw message", "code": 123}'

    with mock.patch.object(errors, expected_error) as error:
        returned = await net.generate_error_response(StubResponse())

    if status_ in (
        http.HTTPStatus.BAD_REQUEST,
        http.HTTPStatus.UNAUTHORIZED,
        http.HTTPStatus.FORBIDDEN,
        http.HTTPStatus.NOT_FOUND,
    ):
        error.assert_called_once_with(
            "https://some.url", {}, '{"message": "raw message", "code": 123}', "raw message", 123
        )
    else:
        error.assert_called_once_with("https://some.url", status_, {}, '{"message": "raw message", "code": 123}')

    assert returned is error()


@pytest.mark.parametrize(
    ("status_", "expected_error"),
    [
        # The following internal server non-conforming status codes are used by cloudflare.
        # Source I made it up...
        # jk https://en.wikipedia.org/wiki/List_of_HTTP_status_codes
        (520, "InternalServerError"),
        (521, "InternalServerError"),
        (522, "InternalServerError"),
        (523, "InternalServerError"),
        (524, "InternalServerError"),
        (525, "InternalServerError"),
        (526, "InternalServerError"),
        (527, "InternalServerError"),
        (530, "InternalServerError"),
        # These non-conforming bad requests status codes are sent by NGINX.
        # Same source as cloudflare status codes.
        (494, "ClientHTTPResponseError"),
        (495, "ClientHTTPResponseError"),
        (496, "ClientHTTPResponseError"),
        (497, "ClientHTTPResponseError"),
        # This non-conforming status code is made up.
        (694, "HTTPResponseError"),
    ],
)
@pytest.mark.asyncio
async def test_generate_error_response_with_non_conforming_status_code(status_, expected_error):
    class StubResponse:
        real_url = "https://some.url"
        status = status_
        headers = {}

        async def read(self):
            return '{"message": "raw message", "code": 123}'

    with mock.patch.object(errors, expected_error) as error:
        returned = await net.generate_error_response(StubResponse())

    error.assert_called_once_with("https://some.url", status_, {}, '{"message": "raw message", "code": 123}')

    assert returned is error()


@pytest.mark.parametrize(
    ("status_", "expected_error"),
    [
        (http.HTTPStatus.UNAUTHORIZED, "UnauthorizedError"),
        (http.HTTPStatus.FORBIDDEN, "ForbiddenError"),
        (http.HTTPStatus.NOT_FOUND, "NotFoundError"),
    ],
)
@pytest.mark.asyncio
async def test_generate_error_when_error_without_json(status_, expected_error):
    class StubResponse:
        real_url = "https://some.url"
        status = status_
        headers = {}

        async def read(self):
            return "some raw body"

    with mock.patch.object(errors, expected_error) as error:
        returned = await net.generate_error_response(StubResponse())

    error.assert_called_once_with("https://some.url", {}, "some raw body")
    assert returned is error()


@pytest.mark.asyncio
async def test_generate_bad_request_error_without_json_response():
    class StubResponse:
        real_url = "https://some.url"
        status = http.HTTPStatus.BAD_REQUEST
        headers = {}
        json = mock.AsyncMock(side_effect=aiohttp.ContentTypeError(None, None))

        async def read(self):
            return "some raw body"

    with mock.patch.object(errors, "BadRequestError", errors=None) as error:
        returned = await net.generate_error_response(StubResponse())

    error.assert_called_once_with("https://some.url", {}, "some raw body", errors=None)
    assert returned is error()


@pytest.mark.parametrize(
    ("data", "expected_errors"),
    [
        ('{"message": "raw message", "code": 123, "errors": {"component": []}}', {"component": []}),
        ('{"message": "raw message", "code": 123, "errors": {}}', {}),
        ('{"message": "raw message", "code": 123}', None),
    ],
)
@pytest.mark.asyncio
async def test_generate_bad_request_error_with_json_response(data, expected_errors):
    class StubResponse:
        real_url = "https://some.url"
        status = http.HTTPStatus.BAD_REQUEST
        headers = {}

        async def read(self):
            return data

    with mock.patch.object(errors, "BadRequestError", errors=None) as error:
        returned = await net.generate_error_response(StubResponse())

    error.assert_called_once_with("https://some.url", {}, data, "raw message", 123, errors=expected_errors)
    assert returned is error()



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import asyncio
import contextlib
import copy
import functools
import inspect
import typing

import async_timeout
import mock

# How long to reasonably expect something to take if it is considered instant.
REASONABLE_QUICK_RESPONSE_TIME = 0.2

# How long to wait for before considering a test to be jammed in an unbreakable
# condition, and thus acceptable to terminate the test and fail it.
REASONABLE_TIMEOUT_AFTER = 5

_T = typing.TypeVar("_T")


def mock_class_namespace(
    klass: type[_T],
    /,
    *,
    init_: bool = True,
    slots_: typing.Optional[bool] = None,
    implement_abstract_methods_: bool = True,
    rename_impl_: bool = True,
    **namespace: typing.Any,
) -> type[_T]:
    """Get a version of a class with the provided namespace fields set as class attributes."""
    if slots_ or slots_ is None and hasattr(klass, "__slots__"):
        namespace["__slots__"] = ()

    if init_ is False:
        namespace["__init__"] = lambda _: None

    if implement_abstract_methods_ and hasattr(klass, "__abstractmethods__"):
        for method_name in klass.__abstractmethods__:
            if method_name in namespace:
                continue

            attr = getattr(klass, method_name)

            if inspect.isdatadescriptor(attr) or inspect.isgetsetdescriptor(attr):
                # Do not use property mock here: it prevents us overwriting it later
                # (e.g. when restubbing for specific test cases)
                namespace[method_name] = mock.Mock(__isabstractmethod__=False)
            elif asyncio.iscoroutinefunction(attr):
                namespace[method_name] = mock.AsyncMock(spec_set=attr, __isabstractmethod__=False)
            else:
                namespace[method_name] = mock.Mock(spec_set=attr, __isabstractmethod__=False)

    for attribute in namespace.keys():
        assert hasattr(klass, attribute), f"invalid namespace attribute {attribute!r} provided"

    name = "Mock" + klass.__name__ if rename_impl_ else klass.__name__

    return type(name, (klass,), namespace)


def retry(max_retries):
    def decorator(func):
        assert asyncio.iscoroutinefunction(func), "retry only supports coroutine functions currently"

        @functools.wraps(func)
        async def retry_wrapper(*args, **kwargs):
            ex = None
            for i in range(max_retries + 1):
                if i:
                    print("retry", i, "of", max_retries)  # noqa: T201 - Print found
                try:
                    await func(*args, **kwargs)
                    return
                except AssertionError as exc:
                    ex = exc  # local variable 'ex' referenced before assignment: wtf?
            raise AssertionError(f"all {max_retries} retries failed") from ex

        return retry_wrapper

    return decorator


def timeout(time_period=REASONABLE_TIMEOUT_AFTER):
    def decorator(func):
        @functools.wraps(func)
        async def wrapper(*args, **kwargs):
            thrown_timeout_error = None

            try:
                async with async_timeout.timeout(time_period):
                    try:
                        await func(*args, **kwargs)
                    except asyncio.TimeoutError as ex:
                        thrown_timeout_error = ex
            except asyncio.TimeoutError as ex:
                raise AssertionError(f"Test took too long (> {time_period}s) and thus failed.") from ex

            if thrown_timeout_error is not None:
                raise thrown_timeout_error

        return wrapper

    return decorator


@contextlib.contextmanager
def ensure_occurs_quickly():
    with async_timeout.timeout(REASONABLE_QUICK_RESPONSE_TIME):
        yield


class ContextManagerMock:
    enter_count = 0
    exit_count = 0

    def __enter__(self):
        self.enter_count += 1
        return self

    def __exit__(self, *args):
        self.exit_count += 1

    def assert_used_once(self):
        assert self.enter_count == 1
        assert self.exit_count == 1


class AsyncContextManagerMock:
    aenter_count = 0
    aexit_count = 0

    async def __aenter__(self):
        self.aenter_count += 1
        return self

    async def __aexit__(self, *args):
        self.aexit_count += 1

    def assert_used_once(self):
        assert self.aenter_count == 1
        assert self.aexit_count == 1


class CopyingAsyncMock(mock.AsyncMock):
    __slots__ = ()

    def __call__(self, *args, **kwargs):
        args = (copy.copy(arg) for arg in args)
        kwargs = {copy.copy(key): copy.copy(value) for key, value in kwargs.items()}
        return super().__call__(*args, **kwargs)



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import enum as py_enum
import os
import sys
import timeit

from hikari.internal import enums as hikari_enum

PyIntFlag = None
HikariIntFlag = None


def build_enums() -> None:
    global PyIntFlag  # noqa: PLW0603
    global HikariIntFlag  # noqa: PLW0603

    class PyIntFlag(py_enum.IntFlag):
        a = 1
        b = 2
        c = 4
        d = 8
        e = 16
        f = 32
        g = 64
        ab = 3
        cde = 28

    class HikariIntFlag(hikari_enum.Flag):
        a = 1
        b = 2
        c = 4
        d = 8
        e = 16
        f = 32
        g = 64
        ab = 3
        cde = 28


if os.name != "nt":
    print("Making highest priority, os.SCHED_RR")
    try:
        pid = os.getpid()
        nice_value = os.nice(-20)
        sys.setswitchinterval(0.5)
        print("sys.getswitchinterval", sys.getswitchinterval())
        os.sched_setaffinity(pid, [(os.cpu_count() or 1) - 1])
        os.sched_setscheduler(pid, os.SCHED_RR, os.sched_param(1))
        print("sched_getscheduler", os.sched_getscheduler(pid))
        print("sched_getparam", os.sched_getparam(pid))
        print("sched_getaffinity", os.sched_getaffinity(pid))
        print("sched_getprioritymax", os.sched_get_priority_max(0))
        print("sched_getprioritymin", os.sched_get_priority_min(0))
        print("sched_rr_getinterval", os.sched_rr_get_interval(pid))
        print("nice", os.nice(0))
    except PermissionError:
        print("run as root to make top OS priority for more accurate results.")
else:
    print("lol windows good luck")

for i in range(5):
    print("pass", i + 1)

    for j in range(1_000_000):
        if sum(j for j in range(10)) < 0:
            raise RuntimeError

    py_intflag_call_time_member = timeit.timeit(
        setup="build_enums()", stmt="PyIntFlag(4)", number=10_000_000, globals=globals()
    )
    hikari_intflag_call_time_member = timeit.timeit(
        setup="build_enums()", stmt="HikariIntFlag(4)", number=10_000_000, globals=globals()
    )

    for j in range(1_000_000):
        if sum(j for j in range(10)) < 0:
            raise RuntimeError

    py_intflag_call_time_existing_composite = (
        timeit.timeit(stmt="PyIntFlag(71)", number=10_000_000, globals=globals()) / 10
    )
    hikari_intflag_call_time_existing_composite = (
        timeit.timeit(stmt="HikariIntFlag(71)", number=10_000_000, globals=globals()) / 10
    )

    for j in range(1_000_000):
        if sum(j for j in range(10)) < 0:
            raise RuntimeError

    build_enums_time = timeit.timeit(stmt="build_enums()", number=10_000, globals=globals())
    py_intflag_call_time_new_composite = timeit.timeit(
        stmt="build_enums(); PyIntFlag(71)", number=10_000, globals=globals()
    )
    build_enums_time = min(timeit.timeit(stmt="build_enums()", number=10_000, globals=globals()), build_enums_time)
    py_intflag_call_time_new_composite -= build_enums_time
    py_intflag_call_time_new_composite *= 100

    for j in range(1_000_000):
        if sum(j for j in range(10)) < 0:
            raise RuntimeError

    build_enums_time = timeit.timeit(stmt="build_enums()", number=10_000, globals=globals())
    hikari_intflag_call_time_new_composite = timeit.timeit(
        stmt="build_enums(); HikariIntFlag(71)", number=10_000, globals=globals()
    )
    build_enums_time = min(timeit.timeit(stmt="build_enums()", number=10_000, globals=globals()), build_enums_time)
    hikari_intflag_call_time_new_composite -= build_enums_time
    hikari_intflag_call_time_new_composite *= 100

    print("PyIntFlag.__call__(4) (existing member)", py_intflag_call_time_member, "µs")
    print("HikariIntFlag.__call__(4) (existing member)", hikari_intflag_call_time_member, "µs")

    print("PyIntFlag.__call__(71) (new composite member)", py_intflag_call_time_new_composite, "µs")
    print("HikariIntFlag.__call__(71) (new composite member)", hikari_intflag_call_time_new_composite, "µs")

    print("PyIntFlag.__call__(71) (existing composite member)", py_intflag_call_time_existing_composite, "µs")
    print("HikariIntFlag.__call__(71) (existing composite member)", hikari_intflag_call_time_existing_composite, "µs")

    print()
